import matplotlib.pyplot as plt
import numpy as np

from shiny import reactive
from shiny.express import input, render, ui

ui.page_opts(title="Shiny Dark Mode", fillable="One")

with ui.nav_panel("One"):
    with ui.layout_sidebar():
        with ui.sidebar():
            ui.input_slider("n", "N", min=0, max=100, value=20)

        @render.plot(alt="A histogram")
        def plot() -> object:
            np.random.seed(19680801)
            x = 100 + 15 * np.random.randn(437)

            fig, ax = plt.subplots()
            ax.hist(x, input.n(), density=True)

            # Theme the plot to match light/dark mode
            fig.patch.set_facecolor("none")
            ax.set_facecolor("none")

            color_fg = "black" if input.mode() == "light" else "silver"
            ax.tick_params(axis="both", colors=color_fg)
            ax.spines["bottom"].set_color(color_fg)
            ax.spines["top"].set_color(color_fg)
            ax.spines["left"].set_color(color_fg)
            ax.spines["right"].set_color(color_fg)

            return fig


with ui.nav_panel("Two"):
    with ui.layout_column_wrap():
        with ui.card():
            "Second page content."

        with ui.card():
            ui.card_header("More content on the second page.")
            ui.input_action_button("make_light", "Switch to light mode")
            ui.input_action_button("make_dark", "Switch to dark mode")

ui.nav_spacer()
with ui.nav_control():
    ui.input_dark_mode(id="mode")


@reactive.effect
@reactive.event(input.make_light)
def _():
    ui.update_dark_mode("light")


@reactive.effect
@reactive.event(input.make_dark)
def _():
    ui.update_dark_mode("dark")



from __future__ import annotations

import ast
from types import CodeType, FunctionType
from typing import Callable

CodeEqual = Callable[[CodeType, CodeType], bool]


def match_name_and_lineno(candidate: CodeType, target: CodeType) -> bool:
    return (
        candidate.co_name == target.co_name
        and candidate.co_firstlineno == target.co_firstlineno
    )


def find_code_for_func(
    code: CodeType, func: FunctionType, equals: CodeEqual = match_name_and_lineno
) -> CodeType | None:
    """
    Given a code object, recurses into it trying to find the code sub-object that is a
    match (for a user-defined definition of "match") for the given function.
    """
    if not isinstance(code, CodeType):
        return None

    if equals(code, func.__code__):
        return code

    for const in code.co_consts:
        if not isinstance(const, CodeType):
            continue

        found = find_code_for_func(const, func, equals)
        if found is not None:
            return found

    return None


def ast_matches_func(node: ast.AST, func: FunctionType) -> bool:
    if not isinstance(node, ast.FunctionDef):
        return False
    linenos = [*[dec.lineno for dec in node.decorator_list], node.lineno]
    return func.__code__.co_firstlineno in linenos and func.__name__ == node.name



from __future__ import annotations

from typing import Any, Callable, List

from htmltools import Tag

from shiny import App, ui
from shiny.types import NavSetArg
from shiny.ui import Sidebar


def nav_controls(prefix: str) -> List[NavSetArg]:
    return [
        ui.nav_panel("a", prefix + ": tab a content"),
        ui.nav_panel("b", prefix + ": tab b content"),
        ui.nav_panel("c", prefix + ": tab c content"),
        ui.nav_spacer(),
        ui.nav_menu(
            "Links",
            ui.nav_control(
                ui.a(
                    "Shiny",
                    href="https://shiny.posit.co/py/",
                    target="_blank",
                )
            ),
            "----",
            "Plain text",
            "----",
            ui.nav_control(
                ui.a(
                    "Posit",
                    href="https://posit.co",
                    target="_blank",
                )
            ),
            align="right",
        ),
    ]


def nav_headerfooter(prefix: str) -> tuple[Tag, Tag]:
    return (
        ui.span(f"{prefix}: Header"),
        ui.span(f"{prefix}: Footer"),
    )


def nav_title(prefix: str) -> Tag:
    return ui.span(f"{prefix}: Title")


def nav_sidebar(prefix: str) -> Sidebar:
    return ui.sidebar(
        f"{prefix}: Sidebar contents",
        title=f"{prefix}: Sidebar Title",
    )


def make_navset(
    name: str,
    fn: Callable[..., Any],
    *,
    controls: bool = True,
    title: bool = False,
    sidebar: bool = False,
    headerfooter: bool = False,
):
    prefix = name + "()"

    args = []
    if controls:
        args = nav_controls(prefix)

    kwargs: dict[str, Any] = {"id": name}
    if title:
        kwargs["title"] = nav_title(prefix)
    if sidebar:
        kwargs["sidebar"] = nav_sidebar(prefix)
    if headerfooter:
        header, footer = nav_headerfooter(prefix)
        kwargs["header"] = header
        kwargs["footer"] = footer

    return ui.TagList(ui.tags.h4(prefix), fn(*args, **kwargs))


app_ui = ui.page_navbar(
    *nav_controls("page_navbar()"),
    # bg="#0062cc",
    # inverse=True,
    id="page_navbar",
    header="page_navbar(): Header",
    fillable=False,
    footer=ui.div(
        {"style": "width:80%;margin: 0 auto"},
        ui.tags.style(
            """
            h4 {
                margin-top: 3em;
            }
            """
        ),
        "page_navbar(): Footer (w/ custom styling)",
        make_navset(
            "navset_bar", ui.navset_bar, title=True, sidebar=True, headerfooter=True
        ),
        make_navset("navset_tab", ui.navset_tab, headerfooter=True),
        make_navset("navset_pill", ui.navset_pill, headerfooter=True),
        make_navset("navset_underline", ui.navset_underline, headerfooter=True),
        make_navset(
            "navset_card_tab",
            ui.navset_card_tab,
            title=True,
            sidebar=True,
            headerfooter=True,
        ),
        make_navset(
            "navset_card_pill",
            ui.navset_card_pill,
            title=True,
            sidebar=True,
            headerfooter=True,
        ),
        make_navset(
            "navset_card_underline",
            ui.navset_card_underline,
            title=True,
            sidebar=True,
            headerfooter=True,
        ),
        make_navset("navset_pill_list", ui.navset_pill_list, headerfooter=True),
    ),
    title=nav_title("page_navbar()"),
    sidebar=nav_sidebar("page_navbar()"),
)


app = App(app_ui, None)



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_selectize("x", "Server side selectize", choices=[], multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        ui.update_selectize(
            "x",
            choices=[f"Foo {i}" for i in range(10000)],
            selected=["Foo 0", "Foo 1"],
            server=True,
        )


app = App(app_ui, server, debug=True)



from __future__ import annotations

import asyncio
import html
import http
import logging
import os
import secrets
import threading
import webbrowser
from typing import TYPE_CHECKING, Callable, Optional, cast

if TYPE_CHECKING:
    import starlette.types
    from asgiref.typing import (
        ASGI3Application,
        ASGIReceiveCallable,
        ASGISendCallable,
        ASGISendEvent,
        HTTPResponseStartEvent,
        Scope,
    )

from ._hostenv import get_proxy_url

# CHILD PROCESS ------------------------------------------------------------


def autoreload_url() -> Optional[str]:
    port = os.getenv("SHINY_AUTORELOAD_PORT")
    if not port:
        return None
    else:
        return get_proxy_url(f"ws://127.0.0.1:{port}/autoreload")


# Instantiated by Uvicorn in worker process when reload is enabled
class HotReloadHandler(logging.Handler):
    """Uvicorn log reader, detects reload events."""

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record: logging.LogRecord) -> None:
        # https://github.com/encode/uvicorn/blob/266db48888f3f1ba56710a49ec82e12eecde3aa3/uvicorn/supervisors/statreload.py#L46
        # https://github.com/encode/uvicorn/blob/266db48888f3f1ba56710a49ec82e12eecde3aa3/uvicorn/supervisors/watchgodreload.py#L148
        if "Reloading..." in record.getMessage():
            reload_begin()
        # https://github.com/encode/uvicorn/blob/926a8f5dc2c9265d5fb5daaafce13878f477e264/uvicorn/lifespan/on.py#L59
        elif "Application startup complete." in record.getMessage():
            reload_end()


# Called from child process when old application instance is shut down
def reload_begin():
    pass


# Called from child process when new application instance starts up
def reload_end():
    import websockets

    # os.kill(os.getppid(), signal.SIGUSR1)

    port = os.getenv("SHINY_AUTORELOAD_PORT")
    if not port:
        return None

    url = f"ws://127.0.0.1:{port}/notify"

    async def _() -> None:
        options = {
            "extra_headers": {
                "Shiny-Autoreload-Secret": os.getenv("SHINY_AUTORELOAD_SECRET", ""),
            }
        }
        try:
            async with websockets.connect(
                url, **options  # pyright: ignore[reportArgumentType]
            ) as websocket:
                await websocket.send("reload_end")
        except websockets.exceptions.ConnectionClosed:
            pass

    asyncio.create_task(_())


class InjectAutoreloadMiddleware:
    """Inserts shiny-autoreload.js into the head.

    It's necessary to do it using middleware instead of in a nice htmldependency,
    because we want autoreload to be effective even when displaying an error page.
    """

    def __init__(
        self,
        app: starlette.types.ASGIApp | ASGI3Application,
        *args: object,
        **kwargs: object,
    ):
        if len(args) > 0 or len(kwargs) > 0:
            raise TypeError(
                f"InjectAutoreloadMiddleware does not support positional or keyword arguments, received {args}, {kwargs}"
            )
        # The starlette types and the asgiref types are compatible, but we'll use the
        # latter internally. See the note in the __call__ method for more details.
        if TYPE_CHECKING:
            app = cast(ASGI3Application, app)
        self.app = app
        ws_url = autoreload_url()
        self.script = (
            f"""  <script src="__shared/shiny-autoreload.js" data-ws-url="{html.escape(ws_url)}"></script>
</head>""".encode(
                "ascii"
            )
            if ws_url
            else bytes()
        )

    async def __call__(
        self,
        scope: starlette.types.Scope | Scope,
        receive: starlette.types.Receive | ASGIReceiveCallable,
        send: starlette.types.Send | ASGISendCallable,
    ) -> None:
        # The starlette types and the asgiref types are compatible, but the latter are
        # more rigorous. In the call interface, we accept both types for compatibility
        # with both. But internally we'll use the more rigorous types.
        # See https://github.com/encode/starlette/blob/39dccd9/docs/middleware.md#type-annotations
        if TYPE_CHECKING:
            scope = cast(Scope, scope)
            receive = cast(ASGIReceiveCallable, receive)
            send = cast(ASGISendCallable, send)

        if scope["type"] != "http":
            return await self.app(scope, receive, send)
        if scope["path"] != "/" or len(self.script) == 0:
            return await self.app(scope, receive, send)

        def mangle_callback(body: bytes) -> tuple[bytes, bool]:
            if b"</head>" in body:
                return (body.replace(b"</head>", self.script, 1), True)
            else:
                return (body, False)

        mangler = ResponseMangler(send, mangle_callback)
        await self.app(scope, receive, mangler.send)


# PARENT PROCESS ------------------------------------------------------------


def start_server(port: int, app_port: int, launch_browser: bool):
    """Starts a websocket server that listens on its own port (separate from the main
    Shiny listener).

    Clients can connect on either the /autoreload or /notify path.

    Clients from the uvicorn worker process connect to the /notify path to notify us
    that a successful startup or reload has occurred.

    Clients from browsers (on localhost only) connect to the /autoreload path to be
    notified when a successful startup or reload has occurred.
    """

    # Store port and secret in environment variables so they are inherited by uvicorn
    # worker processes.
    secret = secrets.token_hex(32)
    os.environ["SHINY_AUTORELOAD_PORT"] = str(port)
    os.environ["SHINY_AUTORELOAD_SECRET"] = secret

    app_url = get_proxy_url(f"http://127.0.0.1:{app_port}/")

    # Run on a background thread so our event loop doesn't interfere with uvicorn.
    # Set daemon=True because we don't want to keep the process alive with this thread.
    threading.Thread(
        None, _thread_main, args=[port, app_url, secret, launch_browser], daemon=True
    ).start()


def _thread_main(port: int, app_url: str, secret: str, launch_browser: bool):
    asyncio.run(_coro_main(port, app_url, secret, launch_browser))


async def _coro_main(
    port: int, app_url: str, secret: str, launch_browser: bool
) -> None:
    import websockets

    reload_now: asyncio.Event = asyncio.Event()

    def nudge():
        nonlocal launch_browser
        if launch_browser:
            # Only launch the browser once, not every time autoreload occurs
            launch_browser = False
            webbrowser.open(app_url, 1)
        reload_now.set()
        reload_now.clear()

    async def reload_server(conn: websockets.server.WebSocketServerProtocol):
        try:
            if conn.path == "/autoreload":
                # The client wants to be notified when the app has reloaded. The client
                # in this case is the web browser, specifically shiny-autoreload.js.
                while True:
                    await reload_now.wait()
                    await conn.send("autoreload")
            elif conn.path == "/notify":
                # The client is notifying us that the app has reloaded. The client in
                # this case is the uvicorn worker process (see reload_end(), above).
                req_secret = conn.request_headers.get("Shiny-Autoreload-Secret", "")
                if req_secret != secret:
                    # The client couldn't prove that they were from a child process
                    return
                data = await conn.recv()
                if isinstance(data, str) and data == "reload_end":
                    nudge()
        except websockets.exceptions.ConnectionClosed:
            pass

    # Handle non-WebSocket requests to the autoreload HTTP server. Without this, if you
    # happen to visit the autoreload endpoint in a browser, you get an error message
    # about only WebSockets being supported. This is not an academic problem as the
    # VSCode extension used in RSW sniffs out ports that are being listened on, which
    # leads to confusion if all you get is an error.
    async def process_request(
        path: str, request_headers: websockets.datastructures.Headers
    ) -> Optional[tuple[http.HTTPStatus, websockets.datastructures.HeadersLike, bytes]]:
        # If there's no Upgrade header, it's not a WebSocket request.
        if request_headers.get("Upgrade") is None:
            # For some unknown reason, this fixes a tendency on GitHub Codespaces to
            # correctly proxy through this request, but give a 404 when the redirect is
            # followed and app_url is requested. With the sleep, both requests tend to
            # succeed reliably.
            await asyncio.sleep(1)
            return (http.HTTPStatus.MOVED_PERMANENTLY, [("Location", app_url)], b"")

    async with websockets.serve(
        reload_server, "127.0.0.1", port, process_request=process_request
    ):
        await asyncio.Future()  # wait forever


class ResponseMangler:
    """A class that assists with intercepting and rewriting response bodies being sent
    over ASGI. This would be easy if not for 1) response bodies are potentially sent in
    chunks, over multiple events; 2) the first response event we receive is the one that
    contains the Content-Length, which can be affected when we do rewriting later on.
    The ResponseMangler handles the buffering and content-length rewriting, leaving the
    caller to only have to worry about the actual body-modifying logic.
    """

    def __init__(
        self, send: ASGISendCallable, mangler: Callable[[bytes], tuple[bytes, bool]]
    ) -> None:
        # The underlying ASGI send function
        self._send = send
        # The caller-provided logic for rewriting the body. Takes a single `bytes`
        # argument that is _all_ of the body bytes seen _so far_, and returns a tuple of
        # (bytes, bool) where the bytes are the (possibly modified) body bytes and the
        # bool is True if the mangler does not care to see any more data.
        self._mangler = mangler

        # If True, the mangler is done and any further data can simply be passed along
        self._done: bool = False

        # Holds the http.response.start event, which may need its Content-Length header
        # rewritten before we send it
        self._response_start: Optional[HTTPResponseStartEvent] = None
        # All the response body bytes we have seen so far
        self._body: bytes = b""

    async def send(self, event: ASGISendEvent) -> None:
        if self._done:
            await self._send(event)
            return

        if event["type"] == "http.response.start":
            self._response_start = event
        elif event["type"] == "http.response.body":
            # This check is mostly to make pyright happy
            if self._response_start is None:
                raise AssertionError(
                    "http.response.body ASGI event sent before http.response.start"
                )

            # Add the newly received body data to what we've seen already
            self._body += event["body"]
            # Snapshot length before we mess with the body
            old_len = len(self._body)
            # Mangle away! If done is True, the mangler doesn't want to do any further
            # mangling.
            self._body, done = self._mangler(self._body)

            new_len = len(self._body)
            if new_len != old_len:
                # The mangling check changed the length of the body. Add the difference
                # to the content-length header (if content-length is even present)
                _add_to_content_length(self._response_start, new_len - old_len)

            more_body = event.get("more_body", False)

            if done or not more_body:
                # Either we've seen the whole body by now (`not more_body`) or the
                # mangler has seen all the data it cares to (`done`). Either way, we can
                # send all the data we have.
                self._done = True
                await self._send(self._response_start)
                await self._send(
                    {
                        "type": "http.response.body",
                        "body": self._body,
                        "more_body": more_body,
                    }
                )
                # Allow gc
                self._response_start = None
                self._body = b""
            else:
                # If we get here, then the mangler isn't done and we are expecting to
                # see more data. Do nothing.
                pass


def _add_to_content_length(event: HTTPResponseStartEvent, offset: int) -> None:
    """If event has a Content-Length header, add the specified number of bytes to it
    (may be negative)"""
    event["headers"] = [
        (
            (name, str(int(value) + offset).encode("latin-1"))
            if name.decode("ascii").lower() == "content-length"
            else (name, value)
        )
        for (name, value) in event["headers"]
    ]



import pandas as pd

from shiny import App, Inputs, reactive, render, ui

app_ui = ui.page_fluid(
    ui.output_data_frame("df1"),
    ui.output_text_verbatim("selected_rows", placeholder=True),
    ui.output_text_verbatim("cell_selection", placeholder=True),
)


def server(input: Inputs):
    df = reactive.Value(pd.DataFrame([[1, 2], [3, 4], [5, 6]], columns=["A", "B"]))

    @render.data_frame
    def df1():
        return render.DataGrid(df(), selection_mode="rows")

    @render.text
    def selected_rows():
        return f"Input selected rows: {input.df1_selected_rows()}"

    @render.text
    def cell_selection():
        return f"Cell selection rows: {df1.cell_selection()['rows']}"


app = App(app_ui, server)



from __future__ import annotations

__all__ = (
    "get_current_session",
    "session_context",
    "require_active_session",
)

from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar

if TYPE_CHECKING:
    from ._session import Session

from .._docstring import no_example
from .._namespaces import namespace_context
from .._typing_extensions import TypedDict


class RenderedDeps(TypedDict):
    deps: list[dict[str, Any]]
    html: str


# ==============================================================================
# Context manager for current session (AKA current reactive domain)
# ==============================================================================
_current_session: ContextVar[Optional[Session]] = ContextVar(
    "current_session", default=None
)
_default_session: Optional[Session] = None


@no_example()
def get_current_session() -> Optional[Session]:
    """
    Get the current user session.

    Returns
    -------
    :
        The current session if one is active, otherwise ``None``.

    Note
    ----
    Shiny apps should not need to call this function directly. Instead, it is intended to
    be used by Shiny developers who wish to create new functions that should only be
    called from within an active Shiny session.

    See Also
    --------
    * :func:`~shiny.session.require_active_session`
    """
    session = _current_session.get()
    return session if session is not None else _default_session


@contextmanager
def session_context(session: Optional[Session]):
    """
    A context manager for current session.

    Parameters
    ----------
    session
        A :class:`~shiny.Session` instance. If not provided, the instance is inferred via
        :func:`~shiny.session.get_current_session`.
    """
    token: Token[Session | None] = _current_session.set(session)
    try:
        with namespace_context(session.ns if session is not None else None):
            yield
    finally:
        _current_session.reset(token)


@no_example()
def require_active_session(session: Optional[Session]) -> Session:
    """
    Raise an exception if no Shiny session is currently active.

    Parameters
    ----------
    session
        A :class:`~shiny.Session` instance. If not provided, the session is inferred via
        :func:`~shiny.session.get_current_session`.

    Returns
    -------
    :
        The session.

    Note
    ----
    Shiny apps should not need to call this function directly. Instead, it is intended to
    be used by Shiny developers who wish to create new functions that should only be
    called from within an active Shiny session.

    Raises
    ------
    ValueError
        If session is not active.

    See Also
    --------
    * :func:`~shiny.session.get_current_session`
    """

    if session is None:
        session = get_current_session()
    if session is None:
        import inspect

        call_stack = inspect.stack()
        if len(call_stack) > 1:
            caller = call_stack[1]
        else:
            # Uncommon case: this function is called from the top-level, so the caller
            # is just require_active_session.
            caller = call_stack[0]

        calling_fn_name = caller.function
        if calling_fn_name == "__init__":
            # If the caller is __init__, then we're most likely in the initialization of
            # an object. This will get the class name.
            calling_fn_name = caller.frame.f_locals["self"].__class__.__name__

        raise RuntimeError(
            f"{calling_fn_name}() must be called from within an active Shiny session."
        )
    return session


# Ideally I'd love not to limit the types for T, but if I don't, the type checker has
# trouble figuring out what `T` is supposed to be when run_thunk is actually used. For
# now, just keep expanding the possible types, as needed.
T = TypeVar("T", str, int)


def read_thunk(thunk: Callable[[], T] | T) -> T:
    if callable(thunk):
        return thunk()
    else:
        return thunk


def read_thunk_opt(thunk: Optional[Callable[[], T] | T]) -> Optional[T]:
    if thunk is None:
        return None
    elif callable(thunk):
        return thunk()
    else:
        return thunk



from shiny import App, Inputs, reactive, render, ui

app_ui = ui.page_fluid(
    ui.input_text("name", "Your Name"),
    ui.input_action_button("greet", "Say Hello", disabled=True),
    ui.output_ui("hello"),
)


def server(input: Inputs):
    @reactive.effect
    @reactive.event(input.name)
    def set_button_state():
        if input.name():
            ui.update_action_button("greet", disabled=False)
        else:
            ui.update_action_button("greet", disabled=True)

    @render.ui
    @reactive.event(input.greet)
    def hello():
        return ui.p(f"Hello, {input.name()}!", class_="fs-1 text-primary mt-3")


app = App(app_ui, server)



from __future__ import annotations

import htmltools

from shiny import App, ui
from shiny.ui import fill


def outer_inner() -> htmltools.Tag:
    inner = ui.div(
        id="inner",
        style=htmltools.css(
            height="200px",
            border="3px blue solid",
        ),
    )
    outer = ui.div(
        inner,
        id="outer",
        style=htmltools.css(
            height="300px",
            border="3px red solid",
        ),
    )
    return outer


outer0 = outer_inner()

outer1 = outer_inner()
outer1 = fill.as_fillable_container(outer1)

outer2 = outer_inner()
outer2 = fill.as_fillable_container(outer2)
outer2.children[0] = fill.as_fill_item(outer2.children[0])

app_ui = ui.page_fluid(
    ui.markdown(
        """\
        # `as_fillable_container()`

        For an item to fill its parent element,
        * the item must have `as_fill_item()` be called on it
        * the parent container must have `as_fillable_container()` called on it

        If both methods are called, the inner child will naturally expand into its parent container.
        """
    ),
    ui.row(
        ui.column(4, ui.h5("Default behavior")),
        ui.column(4, ui.h5(ui.markdown("`as_fillable_container(red)`"))),
        ui.column(
            4,
            ui.h5(ui.markdown("`as_fill_item(blue)` + `as_fillable_container(red)`")),
        ),
    ),
    ui.row(
        ui.column(4, ui.div(outer0)),
        ui.column(4, ui.div(outer1)),
        ui.column(4, ui.span(outer2)),
    ),
)


app = App(app_ui, server=None)



from shiny.express import input, render, ui

ui.input_select(
    "state",
    "Choose a state:",
    {
        "East Coast": {"NY": "New York", "NJ": "New Jersey", "CT": "Connecticut"},
        "West Coast": {"WA": "Washington", "OR": "Oregon", "CA": "California"},
        "Midwest": {"MN": "Minnesota", "WI": "Wisconsin", "IA": "Iowa"},
    },
)


@render.text
def value():
    return "You choose: " + str(input.state())



import pytest
from example_apps import get_apps, reruns, reruns_delay, validate_example
from playwright.sync_api import Page


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("ex_app_path", get_apps("shiny/api-examples"))
def test_api_examples(page: Page, ex_app_path: str) -> None:
    validate_example(page, ex_app_path)



__all__ = ("input_password",)

from typing import Optional

from htmltools import Tag, TagChild, css, div, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._utils import shiny_input_label


@add_example()
def input_password(
    id: str,
    label: TagChild,
    value: str = "",
    *,
    width: Optional[str] = None,
    placeholder: Optional[str] = None,
) -> Tag:
    """
    Create an password control for entry of passwords.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        Initial value.
    width
        The CSS width, e.g., '400px', or '100%'.
    placeholder
        The placeholder of the input.

    Returns
    -------
    :
        A UI element.

    Notes
    ------
    ::: {.callout-note title="Server value"}
    A string of the password input. The default value is unless value is provided.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_text`
    """
    resolved_id = resolve_id(id)
    return div(
        shiny_input_label(resolved_id, label),
        tags.input(
            id=resolved_id,
            type="password",
            value=value,
            class_="shiny-input-password form-control",
            placeholder=placeholder,
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )



import pytest
from playwright.sync_api import Page, expect
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

from shiny.playwright import controller

TIMEOUT = 2 * 60 * 1000

app_url = local_deploys_app_url_fixture("shiny_express_page_default")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_page_default(page: Page, app_url: str) -> None:
    page.goto(app_url)

    # since it is a custom table we can't use the OutputTable controls
    expect(page.locator("#shell")).to_have_text(
        "R1C1R1\nR1C1R2-R1C1R1\nR1C1R2-R1C1R2\nR1C1R2-R1C2\nR1C2", timeout=TIMEOUT
    )

    # Perform these tests second as their locators are not stable over time.
    # (They require that a locator be realized before finding the second locator)
    nav_html = controller.NavsetTab(page, "express_navset_tab")
    nav_html._expect_content_text("pre 0pre 1pre 2")
    nav_html.set("div")
    nav_html._expect_content_text("div 0\ndiv 1\ndiv 2")
    nav_html.set("span")
    nav_html._expect_content_text("span 0span 1span 2")

    navset_card_tab = controller.NavsetTab(page, "express_navset_card_tab")
    navset_card_tab._expect_content_text("")



import matplotlib.pyplot as plt
import numpy as np

from shiny.express import input, render, ui

with ui.sidebar():
    ui.input_slider("n", "N", min=0, max=100, value=20)


@render.plot(alt="A histogram")
def plot() -> object:
    np.random.seed(19680801)
    x = 100 + 15 * np.random.randn(437)

    fig, ax = plt.subplots()
    ax.hist(x, input.n(), density=True)
    return fig



"""
We can't use starlette's StaticFiles when running in wasm mode, because it launches a
thread. Instead, use our own crappy version. Fortunately, this is all we need.

When running in native Python mode, use the starlette StaticFiles impl; it's battle
tested, whereas ours is not. Under wasm, it's OK if ours has bugs, even security holes:
everything is running in the browser sandbox including the filesystem, so there's
nothing we could disclose that an attacker wouldn't already have access to. The same is
not true when running in native Python, we want to be as safe as possible.
"""

from __future__ import annotations

import re
from typing import Any

from . import _utils

__all__ = (
    "StaticFiles",
    "FileResponse",
)

import sys

from starlette.background import BackgroundTask

if "pyodide" not in sys.modules:
    # Running in native mode; use starlette StaticFiles
    import os

    import starlette.responses
    import starlette.staticfiles

    FileResponse = starlette.responses.FileResponse  # type: ignore

    # Wrapper for StaticFiles to fix .js content-type issues on Windows 10 (see #1601)
    class StaticFiles(starlette.staticfiles.StaticFiles):  # type: ignore
        def file_response(
            self,
            full_path: str | os.PathLike[str],
            *args: Any,
            **kwargs: Any,
        ) -> starlette.responses.Response:
            resp = super().file_response(full_path, *args, **kwargs)
            if resp.headers.get("content-type", "").startswith("text/plain"):
                correct_type = _utils.guess_mime_type(full_path)
                resp.headers["content-type"] = (
                    f"{correct_type}; charset={resp.charset}"
                    if correct_type.startswith("text/")
                    else correct_type
                )
                resp.media_type = correct_type
            return resp

else:
    # Running in wasm mode; must use our own simple StaticFiles

    import os
    import os.path
    import pathlib
    import urllib.parse
    from typing import Iterable, MutableMapping, Optional

    from starlette.responses import PlainTextResponse
    from starlette.types import Receive, Scope, Send

    class StaticFiles:
        dir: pathlib.Path
        root_path: str

        def __init__(self, *, directory: str | os.PathLike[str]):
            self.dir = pathlib.Path(os.path.realpath(os.path.normpath(directory)))

        async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
            if scope["type"] != "http":
                raise AssertionError("StaticFiles can't handle non-http request")
            # following starlette >=0.33, tested to be compatible with 0.32-0.37.2
            root_path = scope.get("route_root_path", scope.get("root_path", ""))
            path = scope.get("route_path", re.sub(r"^" + root_path, "", scope["path"]))
            path_segments = path.split("/")
            final_path, trailing_slash = _traverse_url_path(self.dir, path_segments)
            if final_path is None:
                return await Error404()(scope, receive, send)

            if not final_path.exists():
                return await Error404()(scope, receive, send)

            # Sanity check that final path is under self.dir, and if not, 404
            # This construction is like .is_relative_to, but works with Python 3.8
            try:
                final_path.relative_to(self.dir)
            except ValueError:
                return await Error404()(scope, receive, send)

            # Serve up the path

            if final_path.is_dir():
                if trailing_slash:
                    # We could serve up index.html or directory listing if we wanted
                    return await Error404()(scope, receive, send)
                else:
                    # We could redirect with an added "/" if we wanted
                    return await Error404()(scope, receive, send)
            else:
                return await FileResponse(final_path)(scope, receive, send)

    def _traverse_url_path(
        dir: pathlib.Path, path_segments: list[str]
    ) -> tuple[Optional[pathlib.Path], bool]:
        assert len(path_segments) > 0

        new_dir = dir
        path_segment = urllib.parse.unquote(path_segments.pop(0))
        # Gratuitous whitespace is not allowed
        if path_segment != path_segment.strip():
            return None, False

        # Check for illegal paths
        if "/" in path_segment:
            return None, False
        elif path_segment == ".." or path_segment == ".":
            return None, False

        if path_segment != "":
            new_dir = dir / path_segment

        if len(path_segments) == 0:
            return new_dir, path_segment == ""
        else:
            return _traverse_url_path(new_dir, path_segments)

    class Error404(PlainTextResponse):
        def __init__(self):
            super().__init__("404", status_code=404)  # type: ignore

    class FileResponse:
        file: pathlib.Path
        headers: Optional[MutableMapping[str, str]]
        media_type: str

        def __init__(
            self,
            file: pathlib.Path,
            headers: Optional[MutableMapping[str, str]] = None,
            media_type: Optional[str] = None,
            background: Optional[BackgroundTask] = None,
        ) -> None:
            self.headers = headers
            self.file = file
            self.background = background

            if media_type is None:
                media_type = _utils.guess_mime_type(file, strict=False)
            self.media_type = media_type

        async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
            with open(self.file, "rb") as f:
                await send(
                    {
                        "type": "http.response.start",
                        "status": 200,
                        "headers": _convert_headers(self.headers, self.media_type),
                    }
                )

                while True:
                    # In pyodide mode (the only mode in which we use this codepath) the
                    # `send()` callback has quite a bit of per-call overhead, so use a
                    # very large chunk size to keep performance adequate.
                    data = f.read(262144)
                    if len(data) == 0:
                        break
                    await send(
                        {
                            "type": "http.response.body",
                            "body": data,
                            "more_body": True,
                        }
                    )

                await send(
                    {"type": "http.response.body", "body": b"", "more_body": False}
                )
            if self.background:
                await self.background()

    def _convert_headers(
        headers: Optional[MutableMapping[str, str]], media_type: Optional[str] = None
    ) -> Iterable[tuple[bytes, bytes]]:
        if headers is None:
            headers = {}

        header_list = [
            (k.encode("latin-1"), v.encode("latin-1")) for k, v in headers.items()
        ]
        if media_type is not None:
            header_list += [
                (
                    b"Content-Type",
                    media_type.encode("latin-1"),
                )
            ]
        return header_list



from __future__ import annotations

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_dark_mode")


def test_input_dark_mode_follows_system_setting(page: Page, app: ShinyAppProc) -> None:
    page.emulate_media(color_scheme="light")
    page.goto(app.url)

    mode_switch = controller.InputDarkMode(page, "mode")
    mode_switch.expect_mode("light")
    mode_switch.expect_attribute("data-bs-theme")

    page.emulate_media(color_scheme="dark")
    mode_switch = controller.InputDarkMode(page, "mode")
    mode_switch.expect_mode("dark")
    mode_switch.expect_attribute("data-bs-theme")


def test_input_dark_mode_switch(page: Page, app: ShinyAppProc) -> None:
    page.emulate_media(color_scheme="light")
    page.goto(app.url)

    mode_switch = controller.InputDarkMode(page, "mode")
    navbar = controller.NavsetBar(page, "page")
    make_light = controller.InputActionButton(page, "make_light")
    make_dark = controller.InputActionButton(page, "make_dark")

    # Test clicking the dark mode switch
    mode_switch.expect_mode("light").click().expect_mode("dark")

    # Change to nav panel two and trigger server-side changes
    navbar.set("Two")

    make_light.click()
    mode_switch.expect_mode("light")

    make_dark.click()
    mode_switch.expect_mode("dark")



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

slider_app = create_doc_example_core_fixture("input_slider")
template_app = create_doc_example_core_fixture("template")


def test_input_slider_kitchen(page: Page, slider_app: ShinyAppProc) -> None:
    page.goto(slider_app.url)

    # page.set_default_timeout(1000)

    obs = controller.InputSlider(page, "obs")

    expect(obs.loc_label).to_have_text("Number of bins:")

    obs.expect_tick_labels(None)
    obs.expect_value("30")

    obs.expect_animate(False)
    # obs.expect_animate_interval(500)
    # obs.expect_animate_loop(True)
    obs.expect_min("10")
    obs.expect_max("100")
    # obs.expect_from()
    obs.expect_step("1")
    obs.expect_ticks("false")
    obs.expect_sep(",")
    obs.expect_pre(None)
    obs.expect_post(None)
    # obs.expect_data_type()
    obs.expect_time_format(None)
    obs.expect_timezone(None)
    obs.expect_drag_range(None)

    obs.set("42")
    obs.expect_value("42")

    # obs.set_fraction((21.0 - 10.0) / (100.0 - 10.0))
    # obs.expect_value("21")

    # # Duplicate logic of next test. Only difference is `max_err_values=15`
    # try:
    #     obs.set("not-a-number", timeout=800)
    # except ValueError as e:
    #     values_found = '"10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "23", "24", ...'
    #     assert values_found in str(
    #         e
    #     ), "Error message should contain the list of first 15 valid values"

    try:
        obs.set("not-a-number", timeout=800, max_err_values=4)
    except ValueError as e:
        values_found = '"10", "11", "12", "13", ...'
        assert values_found in str(
            e
        ), "Error message should contain the list of first 4 valid values"


def test_input_slider_output(page: Page, template_app: ShinyAppProc) -> None:
    page.goto(template_app.url)

    slider = controller.InputSlider(page, "n")
    txt = controller.OutputTextVerbatim(page, "txt")

    txt.expect_value("n*2 is 40")
    slider.expect_label("N")
    slider.expect_value("20")

    slider.set("42")
    slider.expect_value("42")

    txt.expect_value("n*2 is 84")



"""Tests for `shiny.Session`."""

import pytest

from shiny import ui
from shiny.reactive import effect, flush, isolate
from shiny.session import Inputs
from shiny.types import SilentException


def test_require_active_session_error_messages():
    # require_active_session() should report the caller's name when an error occurs.
    with pytest.raises(RuntimeError, match=r"Progress\(\) must be called"):
        ui.Progress()

    with pytest.raises(RuntimeError, match=r"notification.remove\(\) must be called.*"):
        ui.notification_remove("abc")


def test_input_readonly():
    input = Inputs({})

    with isolate():
        with pytest.raises(RuntimeError):
            input.x.set(1)


def test_input_nonexistent():
    # Make sure that if you try to access an input that doesn't exist, it:
    # - Raises a SilentException
    # - Appears _not_ to add the item to the Inputs object (even though under the hood,
    #   it does actually add it)
    input = Inputs({})

    with isolate():
        assert "x" not in input
        with pytest.raises(SilentException):
            input.x()
        assert "x" not in input
        with pytest.raises(SilentException):
            input.x()

    with isolate():
        with pytest.raises(SilentException):
            input.y()
        assert "y" not in input
        with pytest.raises(SilentException):
            input.y()
        assert "y" not in input


@pytest.mark.asyncio
async def test_input_nonexistent_deps():
    # Make sure that `"x" in input` causes a reactive dependency to be created.
    input = Inputs({})
    result = None

    @effect()
    def o1():
        nonlocal result
        result = "x" in input

    await flush()
    assert result is False
    assert o1._exec_count == 1

    # This should invalidate o1 and cause it to re-execute on the next flush().
    input.x._set(1)
    await flush()
    assert result is True
    assert o1._exec_count == 2

    # This shouldn't invalidate o1, because it doesn't change the status of x's
    # existence. (x already exists; this just changes its value.)
    input.x._set(2)
    await flush()
    assert result is True
    assert o1._exec_count == 2



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_select_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    basic_select = controller.InputSelect(page, "basic_select")
    basic_select_txt = controller.OutputCode(page, "basic_result_txt")
    basic_select.expect_label("Default select")
    basic_select.expect_choices(["Apple", "Banana", "Cherry", "Date", "Elderberry"])
    basic_select.expect_choice_labels(
        ["Apple", "Banana", "Cherry", "Date", "Elderberry"]
    )
    basic_select.expect_choice_groups([])
    basic_select_txt.expect_value("Apple")
    basic_select.expect_multiple(False)

    multiple_select = controller.InputSelect(page, "multi_select")
    multiple_select_txt = controller.OutputCode(page, "multi_result_txt")
    multiple_options = ["Banana", "Cherry"]
    multiple_select.set(multiple_options)
    multiple_select.expect_multiple(True)
    multiple_select_txt.expect_value("Banana, Cherry")

    select_with_selected = controller.InputSelect(page, "select_with_selected")
    select_with_selected_txt = controller.OutputCode(page, "select_with_selected_txt")
    select_with_selected.expect_selected("Cherry")
    select_with_selected_txt.expect_value("Cherry")

    select_with_width = controller.InputSelect(page, "width_select")
    select_with_width.expect_width("400px")

    select_with_labels = controller.InputSelect(page, "select_with_labels")
    select_with_labels.expect_choices(
        ["apple", "banana", "cherry", "date", "elderberry"]
    )
    select_with_labels.expect_choice_labels(
        ["Apple", "Banana", "Cherry", "Date", "Elderberry"]
    )
    controller.OutputCode(page, "select_with_labels_txt").expect_value("apple")

    select_with_custom_size_and_dict = controller.InputSelect(
        page, "select_with_custom_size_and_dict"
    )

    select_with_custom_size_and_dict.expect_choice_groups(["Citrus", "Berries"])
    select_with_custom_size_and_dict.expect_choice_labels(
        [
            "Sweet and tangy",
            "Zesty and refreshing",
            "Bright and tart",
            "Juicy and sweet",
            "Tiny and antioxidant-rich",
            "Delicate and slightly tart",
        ]
    )

    # TODO-karan; Debug why this does not complete
    # page.set_default_timeout(100)
    # select_with_custom_size_and_dict.expect_choices(
    #     [
    #         "Orange",
    #         "Lemon",
    #         "Lime",
    #         "Strawberry",
    #         "Blueberry",
    #         "Raspberry",
    #     ],
    # )
    # select_with_custom_size_and_dict.expect_size("4")

    # # # TODO-karan; Should we implement this? Seems like the only way to determine which choices belong to which groups. While we're at it, we might as well test the labels.
    # # select_with_custom_size_and_dict.expect_choices({
    # #     "Citrus": {
    # #         "Orange": "Sweet and tangy",
    # #         "Lemon": "Zesty and refreshing",
    # #         "Lime": "Bright and tart",
    # #     },
    # #     "Berries": {
    # #         "Strawberry": "Juicy and sweet",
    # #         "Blueberry": "Tiny and antioxidant-rich",
    # #         "Raspberry": "Delicate and slightly tart",
    # #     },
    # # })
    # # This would also allow us to test the labels and values at the same time!
    # # select_with_custom_size_and_dict.expect_choices({
    # #     "Orange": "Sweet and tangy",
    # #     "Lemon": "Zesty and refreshing",
    # #     "Lime": "Bright and tart",
    # #     "Strawberry": "Juicy and sweet",
    # #     "Blueberry": "Tiny and antioxidant-rich",
    # #     "Raspberry": "Delicate and slightly tart",
    # # })
    # ------------------------------------
    # # If we get choices with no labels, should it be auto upgraded to a dictionary of `{FOO: FOO}`?
    # # ANSWER: No, unless we drop `expect_labels()` and `expect_choice_groups()` and just have `expect_choices()` handle everything.
    # # select_with_custom_size_and_dict.expect_choices([
    # #     "Orange",
    # #     "Lemon",
    # #     "Lime",
    # #     "Strawberry",
    # #     "Blueberry",
    # #     "Raspberry",
    # # ])
    # # Auto upgraded to: (which also asserts that there are no groups)
    # # select_with_custom_size_and_dict.expect_choices({
    # #     "Orange": "Orange",
    # #     "Lemon": "Lemon",
    # #     "Lime": "Lime",
    # #     "Strawberry": "Strawberry",
    # #     "Blueberry": "Blueberry",
    # #     "Raspberry": "Raspberry",
    # # })

    # # This should be auto upgraded from
    # # select_with_custom_size_and_dict.expect_choices({
    # #     "Citrus": [
    # #         "Orange",
    # #         "Lemon",
    # #         "Lime",
    # #     ],
    # #     "Berries": [
    # #         "Strawberry",
    # #         "Blueberry",
    # #         "Raspberry",
    # #     ],
    # # })
    # # to:
    # # select_with_custom_size_and_dict.expect_choices({
    # #     "Citrus": {
    # #         "Orange": "Orange",
    # #         "Lemon": "Lemon",
    # #         "Lime": "Lime",
    # #     },
    # #     "Berries": {
    # #         "Strawberry": "Strawberry",
    # #         "Blueberry": "Blueberry",
    # #         "Raspberry": "Raspberry",
    # #     },
    # # })



from pathlib import PurePath

from htmltools import HTMLDependency, Tag

from shiny.module import resolve_id

# This object is used to let Shiny know where the dependencies needed to run
# our component all live. In this case, we're just using a single javascript
# file but we could also include CSS.
custom_component_deps = HTMLDependency(
    "custom_component",
    "1.0.0",
    source={
        "package": "custom_component",
        "subdir": str(PurePath(__file__).parent / "distjs"),
    },
    script={"src": "index.js", "type": "module"},
)


def custom_component(id: str):
    """
    A shiny input.
    """
    return Tag(
        # This is the name of the custom tag we created with our webcomponent
        "custom-component",
        custom_component_deps,
        # Use resolve_id so that our component will work in a module
        id=resolve_id(id),
    )



from shiny.express import render, ui

# Set some Shiny page options
ui.page_opts(title="Hello Chat")

# Create a chat instance, with an initial message
chat = ui.Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
)

# Display the chat
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    user_msg = chat.user_input()
    await chat.append_message(f"You said: {user_msg}")


"Message state:"


@render.code
def message_state():
    return str(chat.messages())



# ------------------------------------------------------------------------------------
# A Shiny Chat example showing how to use different language models via LangChain.
# To run it with all the different providers/models, you'll need API keys for each.
# Namely, OPENAI_API_KEY, ANTHROPIC_API_KEY, and GOOGLE_API_KEY.
# To see how to get these keys, see the relevant basic examples.
# (i.e., ../basic/openai/app.py, ../basic/anthropic/app.py, ../basic/gemini/app.py)
# ------------------------------------------------------------------------------------

from langchain_anthropic import ChatAnthropic
from langchain_google_vertexai import VertexAI
from langchain_openai import ChatOpenAI

from shiny.express import input, render, ui

models = {
    "openai": ["gpt-4o", "gpt-3.5-turbo"],
    "claude": [
        "claude-3-opus-latest",
        "claude-3-5-sonnet-latest",
        "claude-3-haiku-20240307",
    ],
    "google": ["gemini-1.5-pro-latest"],
}

model_choices: dict[str, dict[str, str]] = {}
for key, value in models.items():
    model_choices[key] = dict(zip(value, value))

ui.page_opts(
    title="Shiny Chat Playground",
    fillable=True,
    fillable_mobile=True,
)

with ui.sidebar(position="right"):
    ui.input_select("model", "Model", choices=model_choices)
    ui.input_select(
        "system_actor",
        "Response style",
        choices=["Chuck Norris", "Darth Vader", "Yoda", "Gandalf", "Sherlock Holmes"],
    )
    ui.input_switch("stream", "Stream", value=False)
    ui.input_slider("temperature", "Temperature", min=0, max=2, step=0.1, value=1)
    ui.input_slider("max_tokens", "Max Tokens", min=1, max=4096, step=1, value=100)


@render.express(fill=True, fillable=True)
def chat_ui():
    system_message = {
        "content": f"""
        You are a helpful AI assistant. Provide answers in the style of {input.system_actor()}.
        """,
        "role": "system",
    }
    chat = ui.Chat(id="chat", messages=[system_message])

    model_params = {
        "model": input.model(),
        "temperature": input.temperature(),
        "max_tokens": input.max_tokens(),
    }

    if input.model() in models["openai"]:
        llm = ChatOpenAI(**model_params)
    elif input.model() in models["claude"]:
        llm = ChatAnthropic(**model_params)
    elif input.model() in models["google"]:
        llm = VertexAI(**model_params)
    else:
        raise ValueError(f"Invalid model: {input.model()}")

    @chat.on_user_submit
    async def _():
        messages = chat.messages(format="langchain")
        if input.stream():
            response = llm.astream(messages)
            await chat.append_message_stream(response)
        else:
            response = await llm.ainvoke(messages)
            await chat.append_message(response)

    chat.ui()



from __future__ import annotations

import os
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from brand_yml import Brand
from htmltools import HTMLDependency

from .._versions import bootstrap as v_bootstrap
from ._theme import Theme

YamlScalarType = Union[str, int, bool, float, None]


class ThemeBrandUnmappedFieldError(ValueError):
    def __init__(self, field: str):
        self.field = field
        self.message = f"Unmapped brand.yml field: {field}"
        super().__init__(self.message)

    def __str__(self):
        return self.message


def warn_or_raise_unmapped_variable(unmapped: str):
    if os.environ.get("SHINY_BRAND_YML_RAISE_UNMAPPED") == "true":
        raise ThemeBrandUnmappedFieldError(unmapped)
    else:
        warnings.warn(
            f"Shiny's brand.yml theme does not yet support {unmapped}.",
            stacklevel=4,
        )


color_map: dict[str, list[str]] = {
    # Bootstrap uses $gray-900 and $white for the body bg-color by default, and then
    # swaps them for $gray-100 and $gray-900 in dark mode. brand.yml may end up with
    # light/dark variants for foreground/background, see posit-dev/brand-yml#38.
    "foreground": ["brand--foreground", "body-color", "body-bg-dark"],
    "background": ["brand--background", "body-bg", "body-color-dark"],
    "primary": ["primary"],
    "secondary": ["secondary", "body-secondary-color", "body-secondary"],
    "tertiary": ["body-tertiary-color", "body-tertiary"],
    "success": ["success"],
    "info": ["info"],
    "warning": ["warning"],
    "danger": ["danger"],
    "light": ["light"],
    "dark": ["dark"],
}
"""Maps brand.color fields to Bootstrap Sass variables"""

# https://github.com/twbs/bootstrap/blob/6e1f75/scss/_variables.scss#L38-L49
bootstrap_colors: list[str] = [
    "white",
    "black",
    "blue",
    "indigo",
    "purple",
    "pink",
    "red",
    "orange",
    "yellow",
    "green",
    "teal",
    "cyan",
]
"""
Colors known to Bootstrap

When these colors are named in `colors.palette`, we'll map the brand's colors to the
corresponding Bootstrap color Sass variable.

* [Bootstrap 5 - Colors](https://getbootstrap.com/docs/5.3/customize/color/#color-sass-maps)
"""

# TODO: test that these Sass variables exist in Bootstrap
typography_map: dict[str, dict[str, list[str]]] = {
    "base": {
        "family": ["font-family-base"],
        "size": ["font-size-base"],  # TODO: consider using $font-size-root instead
        "line_height": ["line-height-base"],
        "weight": ["font-weight-base"],
    },
    "headings": {
        "family": ["headings-font-family"],
        "line_height": ["headings-line-height"],
        "weight": ["headings-font-weight"],
        "color": ["headings-color"],
        "style": ["headings-style"],
    },
    "monospace": {
        "family": ["font-family-monospace"],
        "size": ["code-font-size"],
        "weight": ["code-font-weight"],
    },
    "monospace_inline": {
        "family": ["font-family-monospace-inline"],
        "color": ["code-color", "code-color-dark"],
        "background_color": ["code-bg"],
        "size": ["code-inline-font-size"],
        "weight": ["code-inline-font-weight"],
    },
    "monospace_block": {
        "family": ["font-family-monospace-block"],
        "line_height": ["code-block-line-height"],
        "color": ["pre-color"],
        "background_color": ["pre-bg"],
        "weight": ["code-block-font-weight"],
        "size": ["code-block-font-size"],
    },
    "link": {
        "background_color": ["link-bg"],
        "color": ["link-color", "link-color-dark"],
        "weight": ["link-weight"],
        "decoration": ["link-decoration"],
    },
}
"""Maps brand.typography fields to corresponding Bootstrap Sass variables"""


class BrandBootstrapConfigFromYaml:
    """Validate a Bootstrap config from a YAML source"""

    def __init__(
        self,
        path: str,
        version: Any = None,
        preset: Any = None,
        functions: Any = None,
        defaults: Any = None,
        mixins: Any = None,
        rules: Any = None,
    ):

        # TODO: Remove `path` and handle in try/except block in caller
        self._path = path
        self.version = version
        self.preset: str | None = self._validate_str(preset, "preset")
        self.functions: str | None = self._validate_str(functions, "functions")
        self.defaults: dict[str, YamlScalarType] | None = self._validate_defaults(
            defaults
        )
        self.mixins: str | None = self._validate_str(mixins, "mixins")
        self.rules: str | None = self._validate_str(rules, "rules")

    def _validate_str(self, x: Any, param: str) -> str | None:
        if x is None or isinstance(x, str):
            return x

        raise ValueError(
            f"Invalid brand `{self._path}.{param}`. Must be a string or empty."
        )

    def _validate_defaults(self, x: Any) -> dict[str, YamlScalarType] | None:
        if x is None:
            return None

        if not isinstance(x, dict):
            raise ValueError(
                f"Invalid brand `{self._path}.defaults`, must be a dictionary."
            )

        y: dict[Any, Any] = x

        if not all([isinstance(k, str) for k in y.keys()]):
            raise ValueError(
                f"Invalid brand `{self._path}.defaults`, all keys must be strings."
            )

        if not all(
            [v is None or isinstance(v, (str, int, float, bool)) for v in y.values()]
        ):
            raise ValueError(
                f"Invalid brand `{self._path}.defaults`, all values must be scalar."
            )

        res: dict[str, YamlScalarType] = y
        return res


class BrandBootstrapConfig:
    """Convenience class for storing Bootstrap defaults from a brand instance"""

    def __init__(
        self,
        version: Any = v_bootstrap,
        preset: str | None = None,
        functions: str | None = None,
        defaults: dict[str, YamlScalarType] | None = None,
        mixins: str | None = None,
        rules: str | None = None,
    ):
        if not isinstance(version, (str, int)):
            raise ValueError(
                f"Bootstrap version must be a string or integer, not {version!r}."
            )

        v_major = str(version).split(".")[0]
        bs_major = str(v_bootstrap).split(".")[0]

        if v_major != bs_major:
            # TODO (bootstrap-update): Assumes Shiny ships one version of Bootstrap
            warnings.warn(
                f"Shiny does not current support Bootstrap version {v_major}. "
                f"Using Bootstrap v{bs_major} instead.",
                stacklevel=4,
            )
            v_major = bs_major

        self.version = v_major
        self.preset = preset
        self.functions = functions
        self.defaults = defaults
        self.mixins = mixins
        self.rules = rules

    @classmethod
    def from_brand(cls, brand: "Brand"):
        if not brand.defaults:
            return cls()

        shiny_args = {}
        if "shiny" in brand.defaults and "theme" in brand.defaults["shiny"]:
            shiny_args = brand.defaults["shiny"]["theme"]

        shiny = BrandBootstrapConfigFromYaml(
            path="defaults.shiny.theme",
            **shiny_args,
        )

        bs_args = {}
        if "bootstrap" in brand.defaults:
            bs_args = brand.defaults["bootstrap"]

        bootstrap = BrandBootstrapConfigFromYaml(
            path="defaults.bootstrap",
            **bs_args,
        )

        # now combine bootstrap and shiny config options in a way that makes sense
        def join_str(x: str | None, y: str | None):
            return "\n".join([z for z in [x, y] if z is not None])

        defaults: dict[str, YamlScalarType] = {}
        defaults.update(bootstrap.defaults or {})
        defaults.update(shiny.defaults or {})

        return cls(
            version=shiny.version or bootstrap.version or v_bootstrap,
            preset=shiny.preset or bootstrap.preset,
            functions=join_str(bootstrap.functions, shiny.functions),
            defaults=defaults,
            mixins=join_str(bootstrap.mixins, shiny.mixins),
            rules=join_str(bootstrap.rules, shiny.rules),
        )


class ThemeBrand(Theme):
    def __init__(
        self,
        brand: "Brand",
        *,
        include_paths: Optional[str | Path | list[str | Path]] = None,
    ):

        name = self._get_theme_name(brand)
        brand_bootstrap = BrandBootstrapConfig.from_brand(brand)

        # Initialize theme ------------------------------------------------------------
        super().__init__(
            name=name,
            preset=brand_bootstrap.preset,
            include_paths=include_paths,
        )

        self.brand = brand

        # Prep Sass and CSS Variables -------------------------------------------------
        sass_vars_theme_colors, sass_vars_brand_colors, css_vars_brand = (
            ThemeBrand._prepare_color_vars(brand)
        )
        sass_vars_typography = ThemeBrand._prepare_typography_vars(brand)

        # Theme -----------------------------------------------------------------------
        # Defaults are added in reverse order, so each chunk appears above the next
        # layer of defaults. The intended order in the final output is:
        # 1. "Brand" Color palette
        # 2. "Brand" Bootstrap Sass vars
        # 3. "Brand" theme colors
        # 4. "Brand" typography
        # 5. Gray scale variables from "Brand" fg/bg or black/white
        # 6. Fallback vars needed by additional "Brand" rules

        self.add_defaults("", "// *---- brand: end of defaults ----* //", "")
        self._add_sass_ensure_variables()
        self._add_sass_brand_grays()
        self._add_defaults_hdr("typography", **sass_vars_typography)
        self._add_defaults_hdr("theme colors", **sass_vars_theme_colors)
        if brand_bootstrap.defaults:
            self._add_defaults_hdr("bootstrap defaults", **brand_bootstrap.defaults)
        self._add_defaults_hdr("brand colors", **sass_vars_brand_colors)

        # "Brand" rules (now in forwards order)
        self._add_rules_brand_colors(css_vars_brand)
        self._add_sass_brand_rules()
        self._add_brand_bootstrap_other(brand_bootstrap)

    def _get_theme_name(self, brand: "Brand") -> str:
        if not brand.meta or not brand.meta.name:
            return "brand"

        return brand.meta.name.short or brand.meta.name.full or "brand"

    @staticmethod
    def _prepare_color_vars(
        brand: "Brand",
    ) -> tuple[dict[str, str], dict[str, str], list[str]]:
        """Colors: create a dictionary of Sass variables and a list of brand CSS variables"""
        if not brand.color:
            return {}, {}, []

        mapped: dict[str, str] = {}
        brand_sass_vars: dict[str, str] = {}
        brand_css_vars: list[str] = []

        # Map values in colors to their Sass variable counterparts
        for thm_name, thm_color in brand.color.to_dict(include="theme").items():
            if thm_name not in color_map:
                warn_or_raise_unmapped_variable(f"color.{thm_name}")
                continue

            for sass_var in color_map[thm_name]:
                mapped[sass_var] = thm_color

        brand_color_palette = brand.color.to_dict(include="palette")

        # Map the brand color palette to Bootstrap's named colors, e.g. $red, $blue.
        for pal_name, pal_color in brand_color_palette.items():
            if pal_name in bootstrap_colors:
                mapped[pal_name] = pal_color

            # Create Sass and CSS variables for the brand color palette
            # => Sass var: `$brand-{name}: {value}`
            brand_sass_vars.update({f"brand-{pal_name}": pal_color})
            # => CSS var: `--brand-{name}: {value}`
            brand_css_vars.append(f"--brand-{pal_name}: {pal_color};")

        # We keep Sass and "Brand" vars separate so we can ensure "Brand" Sass vars come
        # first in the compiled Sass definitions.
        return mapped, brand_sass_vars, brand_css_vars

    @staticmethod
    def _prepare_typography_vars(brand: "Brand") -> dict[str, str]:
        """Typography: Create a list of Bootstrap Sass variables"""
        mapped: dict[str, str] = {}

        if not brand.typography:
            return mapped

        brand_typography = brand.typography.model_dump(
            exclude={"fonts"},
            exclude_none=True,
            context={"typography_base_size_unit": "rem"},
        )

        for field, prop in brand_typography.items():
            if field not in typography_map:
                warn_or_raise_unmapped_variable(f"typography.{field}")
                continue

            for prop_key, prop_value in prop.items():
                if prop_key in typography_map[field]:
                    typo_sass_vars = typography_map[field][prop_key]
                    for typo_sass_var in typo_sass_vars:
                        mapped[typo_sass_var] = prop_value
                else:
                    warn_or_raise_unmapped_variable(f"typography.{field}.{prop_key}")

        return mapped

    def _add_defaults_hdr(self, header: str, **kwargs: YamlScalarType):
        self.add_defaults(**kwargs)
        self.add_defaults(f"\n// *---- brand: {header} ----* //")

    def _add_sass_ensure_variables(self):
        """Ensure the variables we create to augment Bootstrap's variables exist"""
        self._add_defaults_hdr(
            "added variables",
            **{
                "code-font-weight": None,
                "code-inline-font-weight": None,
                "code-inline-font-size": None,
                "code-block-font-weight": None,
                "code-block-font-size": None,
                "code-block-line-height": None,
                "link-bg": None,
                "link-weight": None,
            },
        )

    def _add_sass_brand_grays(self):
        """
        Adds functions and defaults to handle creating a gray scale palette from the
        brand color palette, or the brand's foreground/background colors.
        """
        self.add_functions(
            """
            @function brand-choose-white-black($foreground, $background) {
              $lum_fg: luminance($foreground);
              $lum_bg: luminance($background);
              $contrast: contrast-ratio($foreground, $background);

              @if $contrast  < 4.5 {
                @warn "The contrast ratio of #{$contrast} between the brand's foreground color (#{inspect($foreground)}) and background color (#{inspect($background)}) is very low. Consider picking colors with higher contrast for better readability.";
              }

              $white: if($lum_fg > $lum_bg, $foreground, $background);
              $black: if($lum_fg <= $lum_bg, $foreground, $background);

              // If the brand foreground/background are close enough to black/white, we
              // use those values. Otherwise, we'll mix the white/black from the brand
              // fg/bg with actual white and black to get something much closer.
              @return (
                "white": if(contrast-ratio($white, white) <= 1.15, $white, mix($white, white, 20%)),
                "black": if(contrast-ratio($black, black) <= 1.15, $black, mix($black, black, 20%)),
              );
            }
            """
        )
        self.add_defaults(
            """
            // *---- brand: automatic gray gradient ----* //
            $enable-brand-grays: true !default;
            // Ensure these variables exist so that we can set them inside of @if context
            // They can still be overwritten by the user, even with !default;
            $white: null !default;
            $black: null !default;
            $gray-100: null !default;
            $gray-200: null !default;
            $gray-300: null !default;
            $gray-400: null !default;
            $gray-500: null !default;
            $gray-600: null !default;
            $gray-700: null !default;
            $gray-800: null !default;
            $gray-900: null !default;

            @if $enable-brand-grays {
              @if variable-exists(brand--foreground) and variable-exists(brand--background) {
                $brand-white-black: brand-choose-white-black($brand--foreground, $brand--background);
                @if $white == null {
                  $white: map-get($brand-white-black, "white") !default;
                }
                @if $black == null {
                  $black: map-get($brand-white-black, "black") !default;
                }
              }
              @if $white != null and $black != null {
                $gray-100: mix($white, $black, 90%) !default;
                $gray-200: mix($white, $black, 80%) !default;
                $gray-300: mix($white, $black, 70%) !default;
                $gray-400: mix($white, $black, 60%) !default;
                $gray-500: mix($white, $black, 50%) !default;
                $gray-600: mix($white, $black, 40%) !default;
                $gray-700: mix($white, $black, 30%) !default;
                $gray-800: mix($white, $black, 20%) !default;
                $gray-900: mix($white, $black, 10%) !default;
              }
            }
            """
        )

    def _add_sass_brand_rules(self):
        """Additional rules to fill in Bootstrap styles for "Brand" parameters"""
        self.add_rules(
            """
            // *---- brand: brand rules to augment Bootstrap rules ----* //
            // https://github.com/twbs/bootstrap/blob/5c2f2e7e/scss/_root.scss#L82
            :root {
                --#{$prefix}link-bg: #{$link-bg};
                --#{$prefix}link-weight: #{$link-weight};
            }
            // https://github.com/twbs/bootstrap/blob/5c2f2e7e/scss/_reboot.scss#L244
            a {
                background-color: var(--#{$prefix}link-bg);
                font-weight: var(--#{$prefix}link-weight);
            }
            code {
                font-weight: $code-font-weight;
            }
            code {
              font-weight: $code-inline-font-weight;
              font-size: $code-inline-font-size;
            }
            // https://github.com/twbs/bootstrap/blob/30e01525/scss/_reboot.scss#L287
            pre {
              font-weight: $code-block-font-weight;
              font-size: $code-block-font-size;
              line-height: $code-block-line-height;
            }

            $bslib-dashboard-design: false !default;
            @if $bslib-dashboard-design and variable-exists(brand--background) {
              // When brand makes dark mode, it usually hides card definition, so we add
              // back card borders in dark mode.
              [data-bs-theme="dark"] {
                --bslib-card-border-color: RGBA(255, 255, 255, 0.15);
              }
            }
            """
        )

    def _add_rules_brand_colors(self, css_vars_colors: list[str]):
        self.add_rules("\n// *---- brand.color.palette ----* //")
        self.add_rules(":root {", *css_vars_colors, "}")

    def _add_brand_bootstrap_other(self, bootstrap: BrandBootstrapConfig):
        if bootstrap.functions:
            self.add_functions(bootstrap.functions)
        if bootstrap.mixins:
            self.add_mixins(bootstrap.mixins)
        if bootstrap.rules:
            self.add_rules(bootstrap.rules)

    def _html_dependencies(self) -> list[HTMLDependency]:
        theme_deps = super()._html_dependencies()

        if not self.brand.typography:
            return theme_deps

        # We're going to put the fonts dependency _inside_ the theme's tempdir, which
        # relies on the theme's dependency having `all_files=True`. We do this because
        # Theme handles the tempdir lifecycle and we want the fonts dependency to be
        # handled in the same way.
        temp_dir = self._get_css_tempdir()
        temp_path = Path(temp_dir) / "fonts"
        temp_path.mkdir(parents=True, exist_ok=True)

        fonts_dep = self.brand.typography.fonts_html_dependency(
            path_dir=temp_path,
            name=f"{self._dep_name()}-fonts",
            version=self._version,
        )

        if fonts_dep is None:
            return theme_deps

        return [fonts_dep, *theme_deps]



from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_text()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.input_text("default", label="Default text input")

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.input_text("placeholder", "With placeholder", placeholder="Enter text here")

        @render.code
        def placeholder_txt():
            return str(input.placeholder())


with ui.layout_columns():
    with ui.card():
        ui.input_text(
            "width", "Custom width", width="200px", value="Custom width input"
        )

        @render.code
        def width_txt():
            return str(input.width())

    with ui.card():
        ui.input_text("autocomplete", "autocomplete input", autocomplete="on")

        @render.code
        def autocomplete_txt():
            return str(input.autocomplete())


with ui.layout_columns():
    with ui.card():
        ui.input_text(
            "spellcheck", "spellcheck input", value="paticular", spellcheck="true"
        )

        @render.code
        def spellcheck_txt():
            return str(input.spellcheck())



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("show", "Show")
ui.input_action_button("remove", "Remove")

ids: list[str] = []
n: int = 0


@reactive.effect
@reactive.event(input.show)
def _():
    global ids
    global n
    # Save the ID for removal later
    id = ui.notification_show("Message " + str(n), duration=None)
    ids.append(id)
    n += 1


@reactive.effect
@reactive.event(input.remove)
def _():
    global ids
    if ids:
        ui.notification_remove(ids.pop())



from __future__ import annotations

import json
import os
import re
import shutil
import sys
import tempfile
import textwrap
import zipfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, Literal, Optional, cast
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import urlopen

import click
import questionary
from questionary import Choice

from ._main_create_custom import (
    ComponentNameValidator,
    update_component_name_in_template,
)
from ._main_utils import (
    cli_action,
    cli_bold,
    cli_code,
    cli_danger,
    cli_field,
    cli_info,
    cli_input,
    cli_success,
    cli_url,
    cli_verbatim,
    cli_wait,
    cli_warning,
    directory_prompt,
)

styles_for_questions = questionary.Style([("secondary", "italic")])
# Prebuild some common choices
cancel_choice: Choice = Choice(title=[("class:secondary", "[Cancel]")], value="cancel")
back_choice: Choice = Choice(title=[("class:secondary", "← Back")], value="back")


def choice_from_templates(templates: list[ShinyTemplate]) -> list[Choice]:
    return [Choice(title=t.title, value=t.id) for t in templates]


@dataclass
class ShinyTemplate:
    """
    Shiny Template

    This class receives metadata for a Shiny template from a `_template.json` file.
    (Alternatively, a template can be created from just `name` and `path` for legacy
    reasons.)

    Attributes
    ----------
    id
        The identifier of the Shiny template. This `id` should be unique within a
        repository of templates.
    path
        The path to the `_template.json` file or the root directory of the template.
    type
        The type of the template (e.g. "app", "package"), default: 'app'.
    title
        A brief title for the template, if provided.
    description
        A longer description of the template, if provided.
    next_steps
        A list of next steps or instructions related to this template, shown after the
        default instructions are displayed. In the `_template.json` file, this field
        can be a single string or an array of strings.
    follow_up
        A list of follow-up actions or information related to this template. In the
        `_template.json` file, this field can be a single string, an array of strings,
        or an object with a `text` field and an optional `type` field. The `type` field
        can be "action", "info", "warning", "danger", or "text".
    """

    id: str
    path: Path
    type: str = "app"
    title: str | None = None
    description: str | None = None
    next_steps: list[str] = field(default_factory=list)
    follow_up: list[ShinyTemplateFollowUp] = field(default_factory=list)
    _express_available: bool | None = None

    @property
    def express_available(self) -> bool:
        """
        Does the template include an Express variant, denoted by the presence of an
        `app-express.py` file?
        """

        if self._express_available is None:
            self._express_available = (self.path / "app-express.py").exists()
        return self._express_available


class ShinyTemplateFollowUp:
    def __init__(
        self,
        text: str,
        type: str = "text",
    ):
        self.text = text
        self.type: Literal["action", "info", "warning", "danger", "text"] = "text"
        if type in ("action", "info", "warning", "danger"):
            self.type = type


def find_templates(path: Path | str = ".") -> list[ShinyTemplate]:
    path = Path(path)
    templates: list[ShinyTemplate] = []
    duplicated_ids: set[str] = set()

    template_files = sorted(path.glob("**/_template.json"))
    for tf in template_files:
        with tf.open() as f:
            try:
                template = json.load(f)
            except json.JSONDecodeError as err:
                raise ValueError(f"Error parsing {tf}: {err}")

            # "next_steps" and "follow_up" can be either a string or an array of strings
            # or an array of dictionaries (follow_up only)
            follow_up_raw: str | list[dict[str, str]] = template.get("follow_up", [])
            if isinstance(follow_up_raw, str):
                follow_up_raw = [{"text": follow_up_raw}]

            follow_up = [ShinyTemplateFollowUp(**f) for f in follow_up_raw]

            next_steps: str | list[str] = template.get("next_steps", [])
            if isinstance(next_steps, str):
                next_steps = [next_steps]

            if "id" not in template:
                raise ValueError(f"Template in {tf} is missing a 'id' field.")

            id = template["id"]
            if id in [t.id for t in templates]:
                duplicated_ids.add(id)

            templates.append(
                ShinyTemplate(
                    id=id,
                    path=tf.parent.absolute(),
                    title=template.get("title"),
                    type=template.get("type", "app"),
                    description=template.get("description"),
                    follow_up=follow_up,
                    next_steps=next_steps,
                )
            )

    if duplicated_ids:
        click.echo(
            cli_danger(
                "Warning: The following templates contain duplicate IDs. "
                + "Only the first occurrence will be used."
            )
        )
        for id in duplicated_ids:
            paths = [t.path.relative_to(path) for t in templates if t.id == id]
            click.echo(
                cli_warning(
                    cli_code(f'"id": "{id}"')
                    + " used by: "
                    + ", ".join([cli_field(str(p)) for p in paths])
                )
            )

    return templates


def template_by_name(templates: list[ShinyTemplate], name: str) -> ShinyTemplate | None:
    for template in templates:
        if template.id == name:
            return template
    return None


class ShinyInternalTemplates:
    """
    Shiny's Internal Templates

    Internal templates that are built into the shiny package are always available via
    `shiny create`. These templates are stored in the `shiny/templates` directory and
    are divided into `app-templates` and `package-templates`.

    To add a new template, create the template subfolder in either of the two template
    folders and add a `_template.json` file. See `ShinyTemplate` for expected fields.

    * `use_template_internal()` is the initial menu seen, which presents `templates/app`
      templates with additional choices.
    * `templates/package` templates are also referred to as `js-components` in the code
      base, these templates appear as a submenu and are handled by
      `use_internal_package_template()`.
    * `templates/chat` templates are generative AI templates and are handled by
      `use_internal_chat_ai_template()`,
    """

    def __init__(self):
        self.templates: dict[str, list[ShinyTemplate]] = {}

    def _templates(self, dir: str = "templates") -> list[ShinyTemplate]:
        if dir in self.templates:
            return self.templates[dir]
        self.templates[dir] = find_templates(Path(__file__).parent / dir)
        return self.templates[dir]

    @property
    def apps(self) -> list[ShinyTemplate]:
        return self._templates("templates/app")

    @property
    def packages(self) -> list[ShinyTemplate]:
        return self._templates("templates/package")

    @property
    def chat_hello_providers(self) -> list[ShinyTemplate]:
        return self._templates("templates/chat/hello-providers")

    @property
    def chat_enterprise(self) -> list[ShinyTemplate]:
        return self._templates("templates/chat/enterprise")

    @property
    def chat_production(self) -> list[ShinyTemplate]:
        return self._templates("templates/chat/production")


shiny_internal_templates = ShinyInternalTemplates()


def use_internal_template(
    question_state: Optional[str] = None,
    mode: Optional[str] = None,
    dest_dir: Optional[Path] = None,
    package_name: Optional[str] = None,
):
    """
    This will initiate a CLI query which will ask the user which template they would like.
    If called without arguments this function will start from the top level and ask which
    type of template the user would like.

    You can also specify a question state to return to another level. For example if you
    were at level 5 of a question chain and wanted to return to level 4.
    This is not that useful currently because we only have two levels of questions.

    :param question_state: The question state you would like to return to. Currently, the options are:
        "cancel": Cancel the operation and exit.
        "js-component": Start the questions for creating a custom JavaScript component.
    """

    app_templates = shiny_internal_templates.apps
    pkg_templates = shiny_internal_templates.packages
    chat_templates = [
        *shiny_internal_templates.chat_hello_providers,
        *shiny_internal_templates.chat_enterprise,
        *shiny_internal_templates.chat_production,
    ]

    menu_choices = [
        Choice(title="Custom JavaScript component...", value="_js-component"),
        Choice(title="Generative AI templates...", value="_chat-ai"),
        Choice(
            title="Choose from the Shiny Templates website", value="_external-gallery"
        ),
    ]

    if question_state is None:
        question_state = question_choose_template(app_templates, *menu_choices)

    template = template_by_name(
        [*app_templates, *pkg_templates, *chat_templates], question_state
    )

    if template is not None:
        if template.type == "app":
            return app_template_questions(template, mode, dest_dir=dest_dir)
        if template.type == "package":
            return package_template_questions(
                template, dest_dir=dest_dir, package_name=package_name
            )

    if question_state == "_external-gallery":
        url = "https://shiny.posit.co/py/templates"
        click.echo(f"Opening {cli_url(url)} in your browser.")
        click.echo(
            f"Choose a template and copy the {cli_code('shiny create')} command to use it."
        )
        import webbrowser

        webbrowser.open(url)
        sys.exit(0)
    elif question_state == "_js-component":
        use_internal_package_template(dest_dir=dest_dir, package_name=package_name)
    elif question_state == "_chat-ai":
        use_internal_chat_ai_template(dest_dir=dest_dir, package_name=package_name)
    else:
        valid_choices = [t.id for t in app_templates + pkg_templates]
        if question_state not in valid_choices:
            raise click.BadOptionUsage(
                "--template",
                f"Invalid value for '--template' / '-t': {question_state} is not one of "
                + f"""'{"', '".join(valid_choices)}'.""",
            )


def use_internal_package_template(
    dest_dir: Optional[Path] = None,
    package_name: Optional[str] = None,
):
    input = questionary.select(
        "What kind of component do you want to build?",
        choices=[
            *choice_from_templates(shiny_internal_templates.packages),
            back_choice,
            cancel_choice,
        ],
        style=styles_for_questions,
    ).ask()

    if input == "back":
        use_internal_template()
        return

    if input is None or input == "cancel":
        sys.exit(1)

    template = template_by_name(shiny_internal_templates.packages, input)

    if template is None:
        # This should be valid because we're selecting from the list of templates
        # but just in case and to make type checkers happy
        raise ValueError(f"Package template for {input} not found.")

    package_template_questions(template, dest_dir=dest_dir, package_name=package_name)


def use_internal_chat_ai_template(
    input: str | None = None,
    dest_dir: Optional[Path] = None,
    package_name: Optional[str] = None,
):
    if input is None:
        input = questionary.select(
            "Which kind of generative AI template would you like to use?",
            choices=[
                Choice(title="By provider...", value="_chat-ai_hello-providers"),
                Choice(title="Enterprise providers...", value="_chat-ai_enterprise"),
                Choice(title="Production-ready chat AI", value="_chat-ai_production"),
                back_choice,
                cancel_choice,
            ],
            style=styles_for_questions,
        ).ask()

        if input is None or input == "cancel":
            sys.exit(1)

        if input == "back":
            use_internal_template(dest_dir=dest_dir, package_name=package_name)
            return

        use_internal_chat_ai_template(
            input, dest_dir=dest_dir, package_name=package_name
        )
        return

    if input == "_chat-ai_production":
        template_choices = shiny_internal_templates.chat_production
    elif input == "_chat-ai_enterprise":
        template_choices = shiny_internal_templates.chat_enterprise
    else:
        template_choices = shiny_internal_templates.chat_hello_providers

    choice = question_choose_template(template_choices, back_choice)

    if choice == "back":
        use_internal_chat_ai_template(dest_dir=dest_dir, package_name=package_name)
        return

    template = template_by_name(
        [
            *shiny_internal_templates.chat_hello_providers,
            *shiny_internal_templates.chat_enterprise,
            *shiny_internal_templates.chat_production,
        ],
        choice,
    )

    if template is None:
        raise ValueError(f"Chat AI template for {choice} not found.")

    app_template_questions(template, dest_dir=dest_dir, mode=None)


def question_choose_template(
    templates: list[ShinyTemplate],
    *extras: Choice,
) -> str:
    """
    Ask the user to pick one of the templates. Includes and handles the cancel choice.
    """

    choice = questionary.select(
        "Which template would you like to use?",
        choices=[*choice_from_templates(templates), *extras, cancel_choice],
        style=styles_for_questions,
    ).ask()

    if choice is None or choice == "cancel":
        sys.exit(1)

    return choice


def download_and_extract_zip(url: str, temp_dir: Path) -> Path:
    try:
        response = urlopen(url)
        data = cast(bytes, response.read())
    except URLError as e:
        # Note that HTTPError is a subclass of URLError
        e.msg += f" for url: {url}"  # pyright: ignore
        raise e

    zip_file_path = temp_dir / "repo.zip"
    zip_file_path.write_bytes(data)
    with zipfile.ZipFile(zip_file_path, "r") as zip_file:
        zip_file.extractall(temp_dir)

    items = list(temp_dir.iterdir())

    # If we unzipped a single directory, return the path to that directory.
    # This avoids much nonsense in trying to guess the directory name, which technically
    # can be derived from the zip file URL, but it's not worth the effort.
    directories = [d for d in items if d.is_dir()]
    files = [f for f in items if f.is_file() and f.name != "repo.zip"]

    # We have exactly one directory and no other files
    if len(directories) == 1 and len(files) == 0:
        return directories[0]

    return temp_dir


def use_github_template(
    github: str,
    template_name: str | None = None,
    mode: str | None = None,
    dest_dir: Path | None = None,
    package_name: str | None = None,
):
    # Github requires that we download the whole repository, so we need to
    # download and unzip the repo, then navigate to the subdirectory.

    spec = parse_github_arg(github)
    spec_cli = f"{spec.repo_owner}/{spec.repo_name}"
    if spec.ref != "HEAD":
        spec_cli = f"{spec_cli}@{spec.ref}"
    if spec.path:
        spec_cli = f"{spec_cli}:{spec.path}"

    click.echo(cli_info(f"Using GitHub repository {cli_field(spec_cli)}."))

    with tempfile.TemporaryDirectory() as temp_dir:
        extracted_dir = None
        errors: list[str] = []

        # Github uses different formats on the `archive/` endpoints for
        # refs/heads/{branch}.zip, refs/tags/{tag}.zip, or just {commit}.zip.
        # It's hard to tell in advance which one is intended, so we try all three in
        # succession, using the first that works.
        for zip_url in github_zip_url(spec):
            try:
                extracted_dir = download_and_extract_zip(zip_url, Path(temp_dir))
                break
            except Exception as err:
                errors.append(str(err))
                pass

        if extracted_dir is None:
            raise click.ClickException(
                f"Failed to download repository from GitHub {cli_url(github)}. "
                + "Please check the URL or GitHub spec and try again.\n"
                + "We received the following errors:\n"
                + textwrap.indent("\n".join(errors), "  ")
            )

        template_dir = extracted_dir / spec.path

        if not os.path.exists(template_dir):
            raise click.ClickException(
                f"Template directory '{cli_input(spec.path)}' does not exist in {cli_field(spec_cli)}."
            )

        templates = find_templates(template_dir)

        if not templates:
            # Legacy: repo doesn't have _template.json files, so we have to rely on
            # paths, i.e. template_dir / template_name
            if template_name is None:
                # warn that we're assuming the repo spec points to the template directly
                click.echo(
                    cli_info(
                        f"Using {cli_field(spec_cli)} as the template. "
                        + f"Use {cli_code('--template')} to specify a template otherwise."
                    )
                )
                template_name = template_dir.name
            else:
                template_dir = template_dir / template_name

            template = ShinyTemplate(
                id=template_name,
                title=f"Template from {spec_cli}",
                path=template_dir,
            )
        elif template_name:
            # Repo has templates and the user already picked one
            template = template_by_name(templates, template_name)
            if not template:
                raise click.ClickException(
                    f"Template '{cli_input(template_name)}' not found in {cli_field(spec_cli)}."
                )
        else:
            # Has templates, but the user needs to pick one
            template_name = question_choose_template(templates)
            template = template_by_name(templates, template_name)

            if not template:
                raise click.ClickException(
                    f"Template '{cli_input(template_name)}' not found in {cli_field(spec_cli)}."
                )

        if template.type == "package":
            return package_template_questions(
                template,
                dest_dir=dest_dir,
                package_name=package_name,
            )
        else:
            return app_template_questions(template, dest_dir=dest_dir, mode=mode)


def github_zip_url(spec: GithubRepoLocation) -> Generator[str]:
    for suffix in ["refs/heads/", "refs/tags/", ""]:
        url = f"https://github.com/{spec.repo_owner}/{spec.repo_name}/archive/{suffix}{spec.ref}.zip"
        yield url


@dataclass
class GithubRepoLocation:
    repo_owner: str
    repo_name: str
    ref: str
    path: str


def parse_github_arg(x: str) -> GithubRepoLocation:
    if re.match(r"^(https?:)//|github[.]com", x):
        return parse_github_url(x)
    return parse_github_spec(x)


def parse_github_spec(spec: str):
    """
    Parses GitHub repo + path spec in the form of:

    * {repo_owner}/{repo_name}@{ref}:{path}
    * {repo_owner}/{repo_name}:{path}@{ref}
    * {repo_owner}/{repo_name}:{path}
    * {repo_owner}/{repo_name}/{path}
    * {repo_owner}/{repo_name}/{path}@{ref}
    * {repo_owner}/{repo_name}/{path}?ref={ref}
    """

    # We split the spec into parts, using a capture group so that the splitting
    # characters are retained in the path parts. Then we know that {repo_owner} /
    # {repo_name} come first in the first three parts, reducing the problem to parsing
    # ref and path in the remaining parts.
    parts = re.split(r"(:|@|[?]ref=|/)", spec)

    if len(parts) < 3:
        raise click.BadParameter(
            f"Could not parse as GitHub spec: '{cli_input(spec)}'"
            + ". Please use the format "
            + cli_field("{repo_owner}/{repo_name}@{ref}:{path}")
            + ".",
            param_hint="--github",
        )

    repo_owner, _, repo_name = parts[:3]
    path = ""
    ref = ""

    which = "path"
    for i, part in enumerate(parts[3:]):
        # special characters indicate a change in which part we're parsing
        if i == 0 and part == "/":
            continue
        if part in ("@", "?ref="):
            which = "ref"
            continue
        if part == ":":
            which = "path"
            continue
        # send each part to the correct repo variable
        if which == "path":
            path += part
        elif which == "ref":
            ref += part

    if ref == "":
        ref = "HEAD"

    return GithubRepoLocation(
        repo_owner=repo_owner,
        repo_name=repo_name,
        ref=ref,
        path=path,
    )


def parse_github_url(x: str) -> GithubRepoLocation:
    """
    Parses a GitHub URL

    e.g. "https://github.com/posit-dev/py-shiny-templates/tree/main/dashboard"
    """

    if not re.match(r"^(https?:)//", x):
        x = "//" + x

    parsed_url = urlparse(x)
    path_parts = parsed_url.path.strip("/").split("/")
    repo_owner, repo_name, _, ref = path_parts[:4]
    path = "/".join(path_parts[4:])
    return GithubRepoLocation(
        repo_owner=repo_owner,
        repo_name=repo_name,
        ref=ref,
        path=path,
    )


def app_template_questions(
    template: ShinyTemplate,
    mode: Optional[str] = None,
    dest_dir: Optional[Path] = None,
):
    template_cli_name = cli_bold(cli_field(template.title or template.id))

    if mode == "express" and not template.express_available:
        raise click.BadParameter(
            f"Express mode not available for the {template_cli_name} template."
        )

    click.echo(cli_wait(f"Creating {template_cli_name} Shiny app..."))
    dest_dir = directory_prompt(dest_dir, template.id)

    if mode is None and template.express_available:
        mode = questionary.select(
            "Would you like to use Shiny Express?",
            [
                Choice("Yes", "express"),
                Choice("No", "core"),
                back_choice,
                cancel_choice,
            ],
        ).ask()

        if mode is None or mode == "cancel":
            sys.exit(1)
        if mode == "back":
            use_internal_template()
            return

    app_dir = copy_template_files(template, dest_dir, mode=mode)

    click.echo(cli_success(f"Created Shiny app at {cli_field(str(app_dir))}"))
    click.echo()
    click.echo(cli_action(cli_bold("Next steps:")))
    if (app_dir / "requirements.txt").exists():
        click.echo("- Install required dependencies:")
        click.echo(
            cli_verbatim(
                [
                    "cd " + str(app_dir) if app_dir != Path(os.curdir) else "",
                    "pip install -r requirements.txt",
                ],
                indent=4,
            )
        )
    click.echo(f"- Open and edit the app file: {cli_field(str(app_dir / 'app.py'))}")

    click_echo_next_steps_and_follow_up(template)


def package_template_questions(
    template: ShinyTemplate,
    dest_dir: Optional[Path] = None,
    package_name: Optional[str] = None,
):
    # Ask what the user wants the name of their component to be
    if package_name is None:
        package_name = questionary.text(
            "What do you want to name your component?",
            instruction="Name must be dash-delimited and all lowercase. E.g. 'my-component-name'",
            validate=ComponentNameValidator,
        ).ask()

        if package_name is None:
            sys.exit(1)

    app_dir = copy_template_files(
        template,
        dest_dir=directory_prompt(dest_dir, package_name),
        mode=None,
    )

    click.echo(cli_wait(f"Setting up {cli_field(package_name)} component package..."))
    update_component_name_in_template(app_dir, package_name)

    click.echo()
    click.echo(cli_action(cli_bold("Next steps:")))
    click.echo("- Setup your component:")
    click.echo(
        cli_verbatim(
            [
                "cd " + str(app_dir) if app_dir != Path(os.curdir) else "",
                "npm install      # install dependencies",
                "npm run build    # build the component",
                "pip install -e . # install the package locally",
            ],
            indent=4,
        )
    )
    click.echo(
        f"- Open and run the example app in the {cli_field('example-app')} directory"
    )

    click_echo_next_steps_and_follow_up(template)


def copy_template_files(
    template: ShinyTemplate,
    dest_dir: Path,
    mode: Optional[str] = None,
):
    files_to_check = [file.name for file in template.path.iterdir()]

    if "__pycache__" in files_to_check:
        files_to_check.remove("__pycache__")

    files_to_check.append("app.py")

    duplicate_files = [file for file in files_to_check if (dest_dir / file).exists()]

    if any(duplicate_files):
        err_files = ", ".join([cli_input('"' + file + '"') for file in duplicate_files])
        click.echo(
            cli_danger(
                "Error: Can't create new files because the following files "
                + f"already exist in the destination directory: {err_files}."
            )
        )
        sys.exit(1)

    if not dest_dir.exists():
        dest_dir.mkdir()

    for item in template.path.iterdir():
        if item.is_file():
            if item.name == "_template.json":
                continue
            shutil.copy(item, dest_dir / item.name)
        else:
            if item.name != "__pycache__":
                shutil.copytree(item, dest_dir / item.name)

    def rename_unlink(file_to_rename: str, file_to_delete: str, dir: Path = dest_dir):
        (dir / file_to_rename).rename(dir / "app.py")
        (dir / file_to_delete).unlink()

    if template.express_available:
        if mode == "express":
            rename_unlink("app-express.py", "app-core.py")
        if mode == "core":
            rename_unlink("app-core.py", "app-express.py")
    if (dest_dir / "app-core.py").exists():
        (dest_dir / "app-core.py").rename(dest_dir / "app.py")

    return dest_dir


def click_echo_next_steps_and_follow_up(template: ShinyTemplate):
    for next_step in template.next_steps:
        click.echo(f"- {next_step}")

    if len(template.follow_up) > 0:
        click.echo()
        for follow_up in template.follow_up:
            click.echo(cli_follow_up(follow_up))


def cli_follow_up(follow_up: ShinyTemplateFollowUp):
    if follow_up.type == "text":
        return follow_up.text
    if follow_up.type == "action":
        return cli_action(follow_up.text)
    if follow_up.type == "info":
        return cli_info(follow_up.text)
    if follow_up.type == "warning":
        return cli_danger(follow_up.text)
    if follow_up.type == "danger":
        return cli_danger(follow_up.text)

    return follow_up.text



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("show", "Show modal dialog"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.show)
    def _():
        m = ui.modal(
            "This is a somewhat important message.",
            title="Somewhat important message",
            easy_close=True,
            footer=None,
        )
        ui.modal_show(m)


app = App(app_ui, server)



from starlette.requests import Request
from starlette.responses import JSONResponse

from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("serve", "Click to serve"), ui.div(id="messages")
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.serve)
    def _():
        async def my_handler(request: Request) -> JSONResponse:
            return JSONResponse({"n_clicks": input.serve()}, status_code=200)

        path = session.dynamic_route("my_handler", my_handler)

        print("Serving at: ", path)

        ui.insert_ui(
            ui.tags.script(
                f"""
                fetch('{path}')
                  .then(r => r.json())
                  .then(x => {{ $('#messages').text(`Clicked ${{x.n_clicks}} times`); }});
                """
            ),
            selector="body",
        )


app = App(app_ui, server)



"""Tools for parsing ipynb files."""

from __future__ import annotations

from pathlib import Path

__all__ = ("convert_code_cells_to_app_py", "get_shiny_deps")

import ast
from typing import Literal, cast

from ._typing_extensions import NotRequired, TypedDict

QuartoShinyCodeCellClass = Literal["python", "r", "cell-code", "hidden"]
QuartoShinyCodeCellContext = Literal["ui", "server", "server-setup"]


class QuartoShinyCodeCell(TypedDict):
    text: str
    context: list[QuartoShinyCodeCellContext]
    classes: list[QuartoShinyCodeCellClass]


class QuartoShinyCodeCells(TypedDict):
    schema_version: int
    cells: list[QuartoShinyCodeCell]
    html_file: str


def convert_code_cells_to_app_py(json_file: str | Path, app_file: str | Path) -> None:
    """Parse an code cell JSON file and output an app.py file."""
    import json
    from textwrap import indent

    json_file = Path(json_file)
    app_file = Path(app_file)

    if app_file.exists():
        with open(app_file, "r", encoding="utf-8") as f:
            first_line = f.readline().strip()
            if first_line != "# This file generated by Quarto; do not edit by hand.":
                raise ValueError(
                    f"Not overwriting app file {app_file}, because it does not appear to be generated by Quarto. "
                    " If this is incorrect, remove the file and try again."
                )

    with open(json_file, "r", encoding="utf-8") as f:
        data = cast(QuartoShinyCodeCells, json.load(f))

    if data["schema_version"] != 1:
        raise ValueError("Only schema_version 1 is supported.")

    cells = data["cells"]

    session_code_cell_texts: list[str] = []
    global_code_cell_texts: list[str] = []

    for cell in cells:
        if "python" not in cell["classes"]:
            continue

        if "server-setup" in cell["context"]:
            global_code_cell_texts.append(cell["text"] + "\n\n# " + "=" * 72 + "\n\n")
        elif "server" in cell["context"]:
            validate_code_has_no_star_import(cell["text"])
            session_code_cell_texts.append(
                indent(cell["text"], "    ") + "\n\n    # " + "=" * 72 + "\n\n"
            )

    app_content = f"""# This file generated by Quarto; do not edit by hand.
# shiny_mode: core

from __future__ import annotations

from pathlib import Path
from shiny import App, Inputs, Outputs, Session, ui

{"".join(global_code_cell_texts)}


def server(input: Inputs, output: Outputs, session: Session) -> None:
{"".join(session_code_cell_texts)}

    return None


_static_assets = ##STATIC_ASSETS_PLACEHOLDER##
_static_assets = {{"/" + sa: Path(__file__).parent / sa for sa in _static_assets}}

app = App(
    Path(__file__).parent / "{data["html_file"]}",
    server,
    static_assets=_static_assets,
)
"""

    with open(app_file, "w", encoding="utf-8") as f:
        f.write(app_content)


# =============================================================================
# HTML Dependency types
# =============================================================================
class QuartoHtmlDepItem(TypedDict):
    name: str
    path: str
    attribs: NotRequired[dict[str, str]]


class QuartoHtmlDepServiceworkerItem(TypedDict):
    source: str
    destination: str


class QuartoHtmlDependency(TypedDict):
    name: str
    version: NotRequired[str]
    scripts: NotRequired[list[str | QuartoHtmlDepItem]]
    stylesheets: NotRequired[list[str | QuartoHtmlDepItem]]
    resources: NotRequired[list[QuartoHtmlDepItem]]
    meta: NotRequired[dict[str, str]]
    serviceworkers: NotRequired[list[QuartoHtmlDepServiceworkerItem]]


def placeholder_dep() -> QuartoHtmlDependency:
    return {
        "name": "shiny-dependency-placeholder",
        "version": "9.9.9",
        "meta": {"shiny-dependency-placeholder": ""},
    }


def get_shiny_deps() -> str:
    import json

    return json.dumps([placeholder_dep()], indent=2)


# =============================================================================
# Functions for checking if code has a star import
# =============================================================================
def validate_code_has_no_star_import(content: str) -> None:
    """
    Check if Python code has a star import at the top level and if so raise an error.

    Parameters
    ----------
    content
        A string with Python code.

    Returns
    -------
    :
        None
    """
    if code_has_star_import(content):
        raise ValueError(
            "'import *' statements cannot be used in a regular Shiny code block in Quarto.\n"
            "Please move '*' imports to a code block with '#| context: setup', or used named imports instead.\n"
        )


def code_has_star_import(content: str) -> bool:
    try:
        tree = ast.parse(content)
        detector = DetectImportStarVisitor()
        detector.visit(tree)

    except Exception:
        return False

    return detector.found_star_import


class DetectImportStarVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.found_star_import = False

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if any(alias.name == "*" for alias in node.names):
            self.found_star_import = True

    # Visit top-level nodes.
    def visit_Module(self, node: ast.Module):
        super().generic_visit(node)

    # Don't recurse into any nodes, so the we'll only ever look at top-level nodes.
    def generic_visit(self, node: ast.AST):
        pass



from __future__ import annotations

import sys
from typing import Any, Callable, TypeVar

import pytest

CallableT = TypeVar("CallableT", bound=Callable[..., Any])


def skip_on_windows(fn: CallableT) -> CallableT:
    fn = pytest.mark.skipif(
        sys.platform.startswith("win"),
        reason="Does not run on windows",
    )(fn)

    return fn



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.playwright import expect as shiny_expect
from shiny.run import ShinyAppProc
from shiny.types import Jsonifiable


def test_validate_column_labels(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    pd_list_styles = controller.OutputDataFrame(page, "pd_list_styles")
    pl_list_styles = controller.OutputDataFrame(page, "pl_list_styles")

    styles: list[dict[str, Jsonifiable]] = [
        {
            "style": {"color": "darkorange", "font-weight": "bold"},
            "class": "everywhere",
        },
        {
            "rows": [1, 2],
            "cols": [2],  # "Species",
            "class": "species",
        },
    ]

    def expect_styles(
        df: controller.OutputDataFrame,
        styles: list[dict[str, Jsonifiable]],
    ):
        def class_for_cell(row: int, col: int) -> list[str]:
            classes: list[str] = []
            for style in styles:
                if "rows" in style:
                    if isinstance(style["rows"], list):
                        if row not in style["rows"]:
                            continue
                if "cols" in style:
                    if isinstance(style["cols"], list):
                        if col not in style["cols"]:
                            continue
                assert isinstance(style["class"], str)
                classes.append(style["class"])
            return classes

        for row in range(0, 5):
            for col in range(0, 6):
                cell_loc = df.cell_locator(row=row, col=col)
                classes = class_for_cell(row, col)

                if len(classes) == 0:
                    shiny_expect.expect_not_to_have_attribute(cell_loc, "class")
                else:
                    for ex_class in classes:
                        assert isinstance(ex_class, str)

                        shiny_expect.expect_to_have_class(cell_loc, ex_class)

    expect_styles(pd_list_styles, styles)
    expect_styles(pl_list_styles, styles)



from shiny.express import input, render, ui

with ui.card(id="my_card", full_screen=True):
    ui.card_header("Hello, card!")
    "A regular full-screenable card."

with ui.value_box(
    id="my_value_box",
    full_screen=True,
    showcase=ui.span("$", class_="fs-1"),
):
    "Hello, value box!"
    "$1,234,567"


@render.code()
def out_card():
    return f"{input.my_card_full_screen()!r}"


@render.code()
def out_value_box():
    return f"{input.my_value_box_full_screen()!r}"



import pathlib

import pandas as pd

from shiny import reactive
from shiny.express import render

file = pathlib.Path(__file__).parent / "mtcars.csv"


@reactive.file_reader(file)
def read_file():
    return pd.read_csv(file)


@render.table
def result():
    return read_file()



from __future__ import annotations

import datetime
import typing

from shiny import App, Inputs, Outputs, Session, render, ui

slider_nums: list[int] = []


def slider_row(
    label: str,
    **kwargs: typing.Any,
) -> ui.Tag:
    id_num = len(slider_nums)
    slider_nums.append(id_num)
    return ui.row(
        ui.column(
            6,
            ui.input_slider(
                f"s{id_num}",
                label,
                **kwargs,
            ),
        ),
        ui.column(
            6,
            ui.output_text_verbatim(f"txt{id_num}", placeholder=True),
        ),
    )


app_ui = ui.page_fluid(
    ui.h2("Sliders!"),
    slider_row(
        "Integer",
        min=0,
        max=1000,
        value=500,
    ),
    slider_row(
        "Range",
        min=1,
        max=1000,
        value=(200, 500),
    ),
    slider_row(
        "Custom Format",
        min=0,
        max=10000,
        value=0,
        step=2500,
        pre="$",
        post=".00",
        sep=",",
        animate=True,
        ticks=True,
    ),
    slider_row(
        "Looping Animation",
        min=1,
        max=2000,
        value=1000,
        step=10,
        animate=ui.AnimationOptions(interval=300, loop=True),
    ),
    slider_row(
        "Single Animation",
        min=0,
        max=5,
        value=1,
        step=1,
        animate=ui.AnimationOptions(interval=100, loop=False),
    ),
    slider_row(
        "Date format",
        min=(datetime.date(2024, 1, 1)),
        max=(datetime.date(2024, 1, 10)),
        value=(datetime.date(2024, 1, 5)),
        time_format="%m/%d/%y",
        timezone="0000",
    ),
    slider_row(
        "Time format",
        min=(datetime.datetime(2024, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)),
        max=(datetime.datetime(2024, 1, 10, 23, 59, tzinfo=datetime.timezone.utc)),
        value=(datetime.datetime(2024, 1, 5, 12, 00, tzinfo=datetime.timezone.utc)),
        width="600px",
        timezone="0000",
    ),
    slider_row(
        "Drag Range (Disabled)",
        min=0,
        max=1000,
        value=(200, 500),
        drag_range=False,
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    def make_output(id_num: int):
        name = f"txt{id_num}"

        @output(id=name)
        @render.text
        def _():
            return input[f"s{id_num}"]()

    for id_num in slider_nums:
        make_output(id_num)


app = App(app_ui, server)



from __future__ import annotations

import asyncio
import contextlib
import functools
import importlib
import inspect
import mimetypes
import os
import random
import secrets
import socketserver
import sys
import tempfile
import warnings
from pathlib import Path
from types import CoroutineType, ModuleType
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Generator,
    Iterable,
    Optional,
    TypeVar,
    cast,
)

from ._typing_extensions import ParamSpec, TypeGuard

CancelledError = asyncio.CancelledError

T = TypeVar("T")


# ==============================================================================
# Misc utility functions
# ==============================================================================
def rand_hex(bytes: int) -> str:
    """
    Creates a random hexadecimal string of size `bytes`. The length in
    characters will be bytes*2.
    """
    format_str = "{{:0{}x}}".format(bytes * 2)
    return format_str.format(secrets.randbits(bytes * 8))


def drop_none(x: dict[str, Any]) -> dict[str, object]:
    return {k: v for k, v in x.items() if v is not None}


# Intended for use with json.load()'s object_hook parameter.
# Note also that object_hook is only called on dicts, not on lists, so this
# won't work for converting "top-level" lists to tuples
def lists_to_tuples(x: object) -> object:
    if isinstance(x, dict):
        x = cast("dict[str, object]", x)
        return {k: lists_to_tuples(v) for k, v in x.items()}
    elif isinstance(x, list):
        x = cast("list[object]", x)
        return tuple(lists_to_tuples(y) for y in x)
    else:
        # TODO: are there other mutable iterators that we want to make read only?
        return x


# Given a dictionary, return a new dictionary with the keys sorted by length.
def sort_keys_length(x: dict[str, T], descending: bool = False) -> dict[str, T]:
    sorted_keys = sorted(x.keys(), key=len, reverse=descending)
    return {key: x[key] for key in sorted_keys}


def guess_mime_type(
    url: "str | os.PathLike[str]",
    default: str = "application/octet-stream",
    strict: bool = True,
) -> str:
    """
    Guess the MIME type of a file. This is a wrapper for mimetypes.guess_type, but it
    only returns the type (and not encoding), and it allows a default value.
    """
    # Note that in the parameters above, "os.PathLike[str]" is in quotes to avoid
    # "TypeError: 'ABCMeta' object is not subscriptable", in Python<=3.8.
    if url:
        # Work around issue #1601, some installations of Windows 10 return text/plain
        # as the mime type for .js files
        _, ext = os.path.splitext(os.fspath(str(url)))
        if ext.lower() in [".js", ".mjs", ".cjs"]:
            return "text/javascript"
    return mimetypes.guess_type(url, strict)[0] or default


def random_port(
    min: int = 1024, max: int = 49151, host: str = "127.0.0.1", n: int = 20
) -> int:
    """Find an open TCP port

    Finds a random available TCP port for listening on, within a specified range
    of ports. The default range of ports to check is 1024 to 49151, which is the
    set of TCP User Ports. This function automatically excludes some ports which
    are considered unsafe by web browsers.

    Parameters
    ----------
    min
        Minimum port number.
    max
        Maximum port number, inclusive.
    host
        Before returning a port number, ensure that we can successfully bind it on this
        host.
    n
        Number of times to attempt before giving up.
    """

    # From https://chromium.googlesource.com/chromium/src.git/+/refs/heads/master/net/base/port_util.cc
    unsafe_ports = [
        1,
        7,
        9,
        11,
        13,
        15,
        17,
        19,
        20,
        21,
        22,
        23,
        25,
        37,
        42,
        43,
        53,
        69,
        77,
        79,
        87,
        95,
        101,
        102,
        103,
        104,
        109,
        110,
        111,
        113,
        115,
        117,
        119,
        123,
        135,
        137,
        139,
        143,
        161,
        179,
        389,
        427,
        465,
        512,
        513,
        514,
        515,
        526,
        530,
        531,
        532,
        540,
        548,
        554,
        556,
        563,
        587,
        601,
        636,
        989,
        990,
        993,
        995,
        1719,
        1720,
        1723,
        2049,
        3659,
        4045,
        5060,
        5061,
        6000,
        6566,
        6665,
        6666,
        6667,
        6668,
        6669,
        6697,
        10080,
    ]

    unusable = set([x for x in unsafe_ports if x >= min and x <= max])
    while n > 0:
        if (max - min + 1) <= len(unusable):
            break
        port = random.randint(min, max)
        if port in unusable:
            continue
        try:
            # See if we can successfully bind
            with socketserver.TCPServer(
                (host, port), socketserver.BaseRequestHandler, bind_and_activate=False
            ) as s:
                s.server_bind()
                return port
        except Exception:
            n -= 1
            continue

    raise RuntimeError("Failed to find a usable random port")


# ==============================================================================
# Private random stream
# ==============================================================================
def private_random_int(min: int, max: int) -> str:
    with private_seed():
        return str(random.randint(min, max))


def private_random_id(prefix: str = "", bytes: int = 3) -> str:
    if prefix != "" and not prefix.endswith("_"):
        prefix += "_"

    with private_seed():
        return prefix + rand_hex(bytes)


@contextlib.contextmanager
def private_seed() -> Generator[None, None, None]:
    state = random.getstate()
    global own_random_state
    try:
        random.setstate(own_random_state)
        yield
    finally:
        own_random_state = random.getstate()
        random.setstate(state)


# Initialize random state for shiny's own private stream of randomness.
current_random_state = random.getstate()
random.seed(secrets.randbits(128))
own_random_state = random.getstate()
random.setstate(current_random_state)

# ==============================================================================
# Async-related functions
# ==============================================================================

R = TypeVar("R")  # Return type
P = ParamSpec("P")


def wrap_async(
    fn: Callable[P, R] | Callable[P, Awaitable[R]]
) -> Callable[P, Awaitable[R]]:
    """
    Given a synchronous function that returns R, return an async function that wraps the
    original function. If the input function is already async, then return it unchanged.
    """

    if is_async_callable(fn):
        return fn

    fn = cast(Callable[P, R], fn)

    @functools.wraps(fn)
    async def fn_async(*args: P.args, **kwargs: P.kwargs) -> R:
        return fn(*args, **kwargs)

    return fn_async


# # TODO-barret-future; Q: Keep code?
# class WrapAsync(Generic[P, R]):
#     """
#     Make a function asynchronous.

#     Parameters
#     ----------
#     fn
#         Function to make asynchronous.

#     Returns
#     -------
#     :
#         Asynchronous function (within the `WrapAsync` instance)
#     """

#     def __init__(self, fn: Callable[P, R] | Callable[P, Awaitable[R]]):
#         if isinstance(fn, WrapAsync):
#             fn = cast(WrapAsync[P, R], fn)
#             return fn
#         self._is_async = is_async_callable(fn)
#         self._fn = wrap_async(fn)

#     async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
#         """
#         Call the asynchronous function.
#         """
#         return await self._fn(*args, **kwargs)

#     @property
#     def is_async(self) -> bool:
#         """
#         Was the original function asynchronous?

#         Returns
#         -------
#         :
#             Whether the original function is asynchronous.
#         """
#         return self._is_async

#     @property
#     def fn(self) -> Callable[P, R] | Callable[P, Awaitable[R]]:
#         """
#         Retrieve the original function

#         Returns
#         -------
#         :
#             Original function supplied to the `WrapAsync` constructor.
#         """
#         return self._fn


# This function should generally be used in this code base instead of
# `iscoroutinefunction()`.
def is_async_callable(
    obj: Callable[P, R] | Callable[P, Awaitable[R]]
) -> TypeGuard[Callable[P, Awaitable[R]]]:
    """
    Determine if an object is an async function.

    This is a more general version of `inspect.iscoroutinefunction()`, which only works
    on functions. This function works on any object that has a `__call__` method, such
    as a class instance.

    Returns
    -------
    :
        Returns True if `obj` is an `async def` function, or if it's an object with a
        `__call__` method which is an `async def` function.
    """
    if inspect.iscoroutinefunction(obj):
        return True
    if hasattr(obj, "__call__"):  # noqa: B004
        if inspect.iscoroutinefunction(obj.__call__):  # type: ignore
            return True

    return False


# def not_is_async_callable(
#     obj: Callable[P, T] | Callable[P, Awaitable[T]]
# ) -> TypeGuard[Callable[P, T]]:
#     return not is_async_callable(obj)


# See https://stackoverflow.com/a/59780868/412655 for an excellent explanation
# of how this stuff works.
# For a more in-depth explanation, see
# https://snarky.ca/how-the-heck-does-async-await-work-in-python-3-5/.
def run_coro_sync(coro: Awaitable[R]) -> R:
    """
    Run a coroutine that is in fact synchronous. Given a coroutine (which is
    returned by calling an `async def` function), this function will run the
    coroutine for one iteration. If the coroutine completes, then return the
    value. If it does not complete, then it will throw a `RuntimeError`.

    What it means to be "in fact synchronous": the coroutine must not yield
    control to the event loop. A coroutine may have an `await` expression in it, and that may call another function that has an `await`, but the chain will
    only yield control if a `yield` statement bubbles through `await`s all the
    way up. For example, `await asyncio.sleep(0)` will have a `yield` which
    bubbles up to the next level. Note that a `yield` in a generator used the
    regular way (not with `await`) will not bubble up, since it is not awaited
    on.
    """
    if not inspect.iscoroutine(coro):
        raise TypeError("run_coro_sync requires a Coroutine object.")

    # Pyright needs a little help here
    coro = cast("CoroutineType[Any, Any, R]", coro)

    try:
        coro.send(None)
    except StopIteration as e:
        return e.value

    raise RuntimeError(
        "async function yielded control; it did not finish in one iteration."
    )


def run_coro_hybrid(coro: Awaitable[R]) -> "asyncio.Future[R]":
    """
    Synchronously runs the given coro up to its first yield, then runs the rest of the
    coro by scheduling it on the current event loop, as per normal. You can think of
    this as either a run_coro_sync() that keeps running in the future, or, as an
    asyncio.create_task() that starts executing immediately instead of via call_soon.

    The status/result/exception can be access through the returned future. Even if an
    error happens synchronously, run_coro_hybrid() will not throw, but rather the error
    will be reported through the future object.

    **PLEASE ONLY USE THIS IF IT'S ABSOLUTELY NECESSARY.** Relative to the official
    asyncio Task implementation, this is a hastily assembled hack job; who knows what
    unknown unknowns lurk here.
    """
    result_future: asyncio.Future[R] = asyncio.Future()

    if not inspect.iscoroutine(coro):
        raise TypeError("run_coro_hybrid requires a Coroutine object.")

    # Pyright needs a little help here
    coro = cast("CoroutineType[Any, Any, R]", coro)

    # Inspired by Task.__step method in cpython/Lib/asyncio/tasks.py
    def _step(fut: Optional["asyncio.Future[None]"] = None):
        assert result_future.cancelled() or not result_future.done()

        exc: Optional[BaseException] = None
        if fut:
            assert fut.done()
            try:
                fut.result()
            except BaseException as e:  # noqa: B036
                exc = e

        if result_future.cancelled():
            # This may cause fut.result()'s exception to be ignored. That's intentional.
            # The cancellation takes precedent, but if we don't call fut.result() first
            # to retrieve its error, Python will warn.
            exc = CancelledError()

        res: Optional[asyncio.Future[None]] = None
        try:
            if exc is None:
                res = coro.send(None)
            else:
                # Is it worth throwing here? Or just logging?
                res = coro.throw(exc)
        except StopIteration as e:
            # Done
            result_future.set_result(e.value)
            return
        except CancelledError:
            result_future.cancel()
            return
        except (KeyboardInterrupt, SystemExit) as e:
            result_future.set_exception(e)
            raise
        except BaseException as e:  # noqa: B036
            result_future.set_exception(e)
        else:
            # If we get here, the coro didn't finish. Schedule it for completion.
            if isinstance(res, asyncio.Future):
                res.add_done_callback(_step)
            elif res is None:
                # This case happens with asyncio.sleep(0)
                asyncio.get_running_loop().call_soon(_step)
            else:
                raise RuntimeError(f"coroutine yielded unknown value: {res!r}")

    _step()

    return result_future


def wrap_async_iterable(x: Iterable[Any] | AsyncIterable[Any]) -> AsyncIterable[Any]:
    """
    Given any iterable, return an async iterable. The async iterable will yield the
    values of the original iterable, but will also yield control to the event loop
    after each value. This is useful when you want to interleave processing with other
    tasks, or when you want to simulate an async iterable from a regular iterable.
    """

    if isinstance(x, AsyncIterable):
        return x

    if not isinstance(x, Iterable):
        raise TypeError("wrap_async_iterable requires an Iterable object.")

    return MakeIterableAsync(x)


class MakeIterableAsync:
    def __init__(self, iterable: Iterable[Any]):
        self.iterable = iterable

    def __aiter__(self):
        self.iterator = iter(self.iterable)
        return self

    async def __anext__(self):
        try:
            value = next(self.iterator)
            await asyncio.sleep(0)  # Yield control to the event loop
            return value
        except StopIteration:
            raise StopAsyncIteration


# ==============================================================================
# Callback registry
# ==============================================================================
class Callbacks:
    def __init__(self) -> None:
        self._callbacks: dict[int, tuple[Callable[[], None], bool]] = {}
        self._id: int = 0

    def register(
        self, fn: Callable[[], None], once: bool = False
    ) -> Callable[[], None]:
        self._id += 1
        id = self._id
        self._callbacks[id] = (fn, once)

        def _():
            if id in self._callbacks:
                del self._callbacks[id]

        return _

    def invoke(self) -> None:
        # The list() wrapper is necessary to force collection of all the items before
        # iteration begins. This is necessary because self._callbacks may be mutated
        # by callbacks.
        for id, value in list(self._callbacks.items()):
            fn, once = value
            try:
                fn()
            finally:
                if once:
                    if id in self._callbacks:
                        del self._callbacks[id]

    def count(self) -> int:
        return len(self._callbacks)


class AsyncCallbacks:
    def __init__(self) -> None:
        self._callbacks: dict[int, tuple[Callable[[], Awaitable[None]], bool]] = {}
        self._id: int = 0

    def register(
        self, fn: Callable[[], Awaitable[None]], once: bool = False
    ) -> Callable[[], None]:
        self._id += 1
        id = self._id
        self._callbacks[id] = (fn, once)

        def _():
            if id in self._callbacks:
                del self._callbacks[id]

        return _

    async def invoke(self) -> None:
        # The list() wrapper is necessary to force collection of all the items before
        # iteration begins. This is necessary because self._callbacks may be mutated
        # by callbacks.
        for id, value in list(self._callbacks.items()):
            fn, once = value
            try:
                await fn()
            finally:
                if once:
                    if id in self._callbacks:
                        del self._callbacks[id]

    def count(self) -> int:
        return len(self._callbacks)


# ==============================================================================
# System-related functions
# ==============================================================================


# Return directory that a package lives in.
def package_dir(package: str) -> str:
    with tempfile.TemporaryDirectory():
        pkg_file = importlib.import_module(".", package=package).__file__
        if pkg_file is None:
            raise RuntimeError(f"Could not find package dir for '{package}'")
        return os.path.dirname(pkg_file)


class ModuleImportWarning(ImportWarning):
    pass


warnings.simplefilter("always", ModuleImportWarning)


def import_module_from_path(module_name: str, path: Path):
    import importlib.util

    if not path.is_absolute():
        raise ValueError("Path must be absolute")

    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not import module {module_name} from path: {path}")

    module = importlib.util.module_from_spec(spec)

    prev_module: ModuleType | None = None

    if module_name in sys.modules:
        prev_module = sys.modules[module_name]
        warnings.warn(
            f"A module named {module_name} is already loaded, but is being loaded again.",
            ModuleImportWarning,
            stacklevel=1,
        )

    sys.modules[module_name] = module

    try:
        spec.loader.exec_module(module)
    except Exception:
        if prev_module is None:
            del sys.modules[module_name]
        else:
            sys.modules[module_name] = prev_module
        raise
    return module



from __future__ import annotations

from typing import Callable, List

from ...types import ListOrTuple
from ._tbl_data import as_data_frame
from ._types import BrowserStyleInfo, IntoDataFrameT, StyleInfo

StyleFn = Callable[[IntoDataFrameT], List["StyleInfo"]]


def style_info_to_browser_style_info(
    info: StyleInfo,
    *,
    nrow: int,
    browser_column_names: ListOrTuple[str],
) -> BrowserStyleInfo | None:
    if not isinstance(info, dict):
        raise TypeError("`StyleInfo` objects must be a dictionary. Received: ", info)

    location = info.get("location", None)
    if location is None:
        location = "body"

    if location != "body":
        raise ValueError(
            f"`StyleInfo` `location` value must be 'body', not '{location}'"
        )

    rows = style_info_rows(info, nrow=nrow)
    cols = style_info_cols(info, browser_column_names=browser_column_names)

    style = info.get("style", None)
    if style is not None:
        if isinstance(style, str):
            raise TypeError(
                "`StyleInfo` `style` value must be a dictionary of key-value pairs, not a style string"
            )
        assert isinstance(style, dict)

    class_ = info.get("class", None)
    if class_ is not None:
        assert isinstance(class_, str), "`StyleInfo` `class` value must be a string"

    if style is None and class_ is None:
        return None

    return {
        "location": location,
        "rows": rows,
        "cols": cols,
        "style": style,
        "class": class_,
    }


def style_info_cols(
    info: StyleInfo,
    *,
    browser_column_names: ListOrTuple[str],
) -> tuple[int, ...] | None:
    cols = info.get("cols", None)
    if cols is None:
        return None
    if isinstance(cols, (str, int)):
        cols = (cols,)
    if not isinstance(cols, (list, tuple)):
        raise TypeError(
            "`StyleInfo` `cols` value must be a list, tuple, int, or string"
        )

    # cols_tup = tuple(cols)
    if len(cols) == 0:
        return ()

    if isinstance(cols[0], bool):
        if len(cols) != len(browser_column_names):
            raise ValueError(
                f"Style column boolean list must have the same length as the data frame column names. Expected {len(browser_column_names)}, got {len(cols)}"
            )
        return tuple(i for i, col in enumerate(cols) if col)

    if isinstance(cols[0], str):
        ret: list[int] = []
        col_names = set(browser_column_names)
        for col in cols:
            assert isinstance(
                col, str
            ), "All elements of `StyleInfo` `cols` must be of the same type: str, int, or bool."
            if col not in col_names:
                raise ValueError(
                    f"`StyleInfo` `cols` value '{col}' not found in data frame column names"
                )
            ret.append(browser_column_names.index(col))
        return tuple(ret)
    elif isinstance(cols[0], int):
        ret: list[int] = []
        for col in cols:
            assert isinstance(
                col, int
            ), "All elements of `StyleInfo` `cols` must be of the same type: str, int, or bool."
            if col < 0:
                raise ValueError(f"Style column index `{col}` must be non-negative")
            if col >= len(browser_column_names):
                raise ValueError(
                    f"Style column index `{col}` out of range (max: {len(browser_column_names) - 1})"
                )
            ret.append(col)
        return tuple(ret)
    elif isinstance(cols[0], bool):
        if all(isinstance(col, bool) for col in cols):
            return tuple(i for i, val in enumerate(cols) if val)
        raise TypeError(
            "All elements of `StyleInfo` `cols` must be of the same type: str, int, or bool."
        )
    else:
        raise TypeError(
            "All elements of `StyleInfo` `cols` must be of the same type: str, int, or bool."
        )


def style_info_rows(
    info: StyleInfo,
    *,
    nrow: int,
) -> None | tuple[int, ...]:
    rows = info.get("rows", None)
    if rows is None:
        return None
    if isinstance(rows, (bool, int)):
        return (rows,)
    if not isinstance(rows, (list, tuple)):
        raise TypeError("`StyleInfo` `rows` value must be a list, tuple, int or string")

    rows_tup = tuple(rows)
    if len(rows_tup) == 0:
        return rows_tup

    if isinstance(rows_tup[0], bool):
        assert (
            len(rows_tup) == nrow
        ), "Length of `StyleInfo` `rows` must match the number of rows in the data frame when `rows` is a boolean list / tuple."
        if all(isinstance(row, bool) for row in rows_tup):
            # Turn into a tuple of indices
            ret = tuple(i for i, val in enumerate(rows_tup) if val)
            return ret

        raise TypeError(
            "All elements of `StyleInfo` `rows` must be of the same type: bool or int."
        )

    elif isinstance(rows_tup[0], int):
        if all(isinstance(row, int) for row in rows_tup):
            return rows_tup

        raise TypeError(
            "All elements of `StyleInfo` `rows` must be of the same type: bool or int."
        )
    else:
        raise TypeError(
            "`StyleInfo` `rows` value must be a list or tuple containing ints, bools, or str values."
        )


def as_style_infos(
    infos: StyleInfo | list[StyleInfo] | StyleFn[IntoDataFrameT] | None,
) -> list[StyleInfo] | StyleFn[IntoDataFrameT]:

    if callable(infos):
        return infos

    if not infos:
        return []

    if not isinstance(infos, list):
        infos = [infos]

    return infos


def as_browser_style_infos(
    infos: list[StyleInfo] | StyleFn[IntoDataFrameT],
    *,
    into_data: IntoDataFrameT,
) -> list[BrowserStyleInfo]:

    if callable(infos):
        style_infos = infos(into_data)
    else:
        style_infos = infos

    if not style_infos:
        return []

    if not isinstance(style_infos, list):
        style_infos = [style_infos]

    nw_data = as_data_frame(into_data)
    browser_column_names = nw_data.columns
    nrow = nw_data.shape[0]

    browser_infos = [
        style_info_to_browser_style_info(
            info,
            nrow=nrow,
            browser_column_names=browser_column_names,
        )
        for info in style_infos
    ]
    return [browser_info for browser_info in browser_infos if browser_info is not None]



import asyncio

from shiny import reactive
from shiny.express import input, render, ui

ui.input_action_button("button", "Compute")


@render.text
@reactive.event(input.button)
async def compute():
    with ui.Progress(min=1, max=15) as p:
        p.set(message="Calculation in progress", detail="This may take a while...")

        for i in range(1, 15):
            p.set(i, message="Computing")
            await asyncio.sleep(0.1)
            # Normally use time.sleep() instead, but it doesn't yet work in Pyodide.
            # https://github.com/pyodide/pyodide/issues/2354

    return "Done computing!"



"""Facade classes for working with Shiny inputs/outputs in Playwright"""

from __future__ import annotations

import re
import sys

from playwright.sync_api import Locator
from playwright.sync_api import expect as playwright_expect

from ..._docstring import no_example
from ..._typing_extensions import assert_type
from .._types import PatternOrStr, PatternStr, Timeout

__all__ = (
    "expect_not_to_have_attribute",
    "expect_to_have_class",
    "expect_not_to_have_class",
    "expect_to_have_style",
    "expect_not_to_have_style",
)


@no_example()
def expect_not_to_have_attribute(
    loc: Locator,
    name: str,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect that the attribute does not exist.

    This method wraps Playwright's Locator expectation `not_to_have_attribute()` and
    sets the value to a regular expression that matches anything.

    If you'd like to check for a specific value, use Playwright's
    `expect(loc).to_have_attribute()` method, instead.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    name
        The attribute name that should **not** be found within the resolved locator.
    timeout
        The maximum time to wait for the attribute to appear
    """
    # Not allowed to have any value for the attribute
    playwright_expect(loc).not_to_have_attribute(
        name, re.compile(r".*"), timeout=timeout
    )
    return


@no_example()
def expect_to_have_class(
    loc: Locator,
    class_: str,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect a CSS class value is found.

    This method wraps Playwright's Locator expectation [`to_have_class()`
    method](https://playwright.dev/python/docs/api/class-locatorassertions#locator-assertions-to-have-class).
    However, Playwright does not have an easy method to check for individual class
    values _within_ the elements `class` value.

    This method will insert the class value
    into a regex pattern to check for the class value within the `class` attribute
    according to word boundaries or the start/end of the `class` string.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    class_
        The class value to find.
    timeout
        The maximum time to wait for the class to appear.

    See Also
    --------
    * :func:`shiny.playwright.expect.expect_not_to_have_class`
    """
    cls_regex = re.compile(rf"(^|\s+){re.escape(class_)}(\s+|$)")
    playwright_expect(loc).to_have_class(cls_regex, timeout=timeout)


@no_example()
def expect_not_to_have_class(
    loc: Locator,
    class_: str,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect a CSS class value is not found.

    This method wraps Playwright's Locator expectation [`not_to_have_class()`
    method](https://playwright.dev/python/docs/api/class-locatorassertions#locator-assertions-not-to-have-class).
    However, Playwright does not have an easy method to check for individual class
    values _within_ the elements `class` value.

    This method will insert the class string
    value into a regex pattern to check for the class value within the `class` attribute
    according to word boundaries or the start/end of the `class` string.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    class_
        The class value that should **not** be found within the resolved locator.
    timeout
        The maximum time to wait for the class to disappear.

    See Also
    --------
    * :func:`shiny.playwright.expect.expect_to_have_class`
    """
    cls_regex = re.compile(rf"(^|\s+){re.escape(class_)}(\s+|$)")
    playwright_expect(loc).not_to_have_class(cls_regex, timeout=timeout)


@no_example()
def expect_to_have_style(
    loc: Locator,
    key: str,
    # Str representation for value. Will be put in a regex with `css_key`
    value: PatternOrStr,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect the `style` attribute to have a value.

    This is different than Playwright's [`to_have_css(key, value)`
    method](https://playwright.dev/python/docs/api/class-locatorassertions#locator-assertions-to-have-css),
    as that method will check for the computed style of the element. Whereas this method
    will check the `style` attribute directly.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    key
        The CSS key to check.
    value
        The CSS value to check.
    timeout
        The maximum time to wait for the style to appear.

    See Also
    --------
    * :func:`shiny.playwright.expect.expect_not_to_have_style`
    """

    playwright_expect(loc).to_have_attribute(
        "style",
        _style_match_str(key, value),
        timeout=timeout,
    )


@no_example()
def expect_not_to_have_style(
    loc: Locator,
    key: str,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect a key within `style` attribute to not exist.

    Convenience method to check if a CSS key does not exist within the `style`
    attribute. The `key` is escaped within a regular expression that checks within
    word boundaries or the start/end of the `style` string.

    This is different than Playwright's [`to_have_css(key, value)`
    method](https://playwright.dev/python/docs/api/class-locatorassertions#locator-assertions-to-have-css),
    as that method will check for the computed style of the element. Whereas this method
    will check the `style` attribute directly.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    key
        The CSS key to check.
    timeout
        The maximum time to wait for the style to appear.

    See Also
    --------
    * :func:`shiny.playwright.expect.expect_to_have_style
    """
    # Not allowed to have any value for the style
    playwright_expect(loc).not_to_have_attribute(
        "style",
        re.compile(rf"\b{re.escape(key)}\s*:"),
        timeout=timeout,
    )
    return


def _attr_match_str(key: str, value: str) -> str:
    # Escape double quotes
    value_str = value.replace('"', '\\"')
    # `key` is `value`
    return f'{key}="{value_str}"'
    # assert_type(value, re.Pattern[str])
    # # `key` contains `value`
    # return f'{key}*="{value.pattern}"'


def _xpath_match_str(key: str, value: PatternOrStr) -> str:
    if isinstance(value, str):
        # Escape double quotes
        value_str = value.replace('"', '\\"')
        # `key` is `value`
        return f'@{key}="{value_str}"'
    else:
        # Disabling type assertion for earlier versions of Python
        if sys.version_info >= (3, 10):
            assert_type(value, re.Pattern[str])

        # `key` contains `value`
        return f'matches(@{key}, "{value.pattern}")'


def _style_match_str(key: str, value: PatternOrStr) -> PatternStr:
    if isinstance(value, str):
        value_str = re.escape(value)
    else:
        value_str = value.pattern
    return re.compile(rf"(^|;)\s*{re.escape(key)}\s*:\s*{value_str}\s*(;|$)")



try:
    import playwright  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]
except ImportError:
    raise ImportError(
        "The shiny.playwright module requires the playwright package to be installed."
        " Please install it with this command:"
        "\n\n    pip install playwright"
    )
# If `pytest` is installed...
try:
    import pytest  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]

    # At this point, `playwright` and `pytest` are installed.
    # Try to make sure `pytest-playwright` is installed
    try:
        import pytest_playwright  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]

    except ImportError:
        raise ImportError(
            "If you are using pytest to test your app,"
            " you can install the pytest-playwright shim package with this command:",
            "\n\n    pip install pytest-playwright",
        )
except ImportError:
    pass
from . import controller, expect

__all__ = ["expect", "controller"]



from playwright.sync_api import Page, expect

from shiny.run import ShinyAppProc


def test_implicit_register(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    expect(page.locator("#out2")).to_have_text("One")
    expect(page.locator("#out3")).to_have_text("Two")
    expect(page.locator("#out4")).to_have_text("Two")
    expect(page.locator("#out1")).to_be_empty()



from __future__ import annotations

__all__ = ("current_namespace", "resolve_id", "ui", "server", "ResolvedId")

from typing import TYPE_CHECKING, Callable, TypeVar

from ._docstring import no_example
from ._namespaces import (
    Id,
    ResolvedId,
    current_namespace,
    namespace_context,
    resolve_id,
)
from ._typing_extensions import Concatenate, ParamSpec

if TYPE_CHECKING:
    from .session import Inputs, Outputs, Session

P = ParamSpec("P")
R = TypeVar("R")

# Ensure that Id type is not stripped out from .pyi file when generating type stubs
_: Id  # type: ignore


@no_example()
def ui(fn: Callable[P, R]) -> Callable[Concatenate[str, P], R]:
    def wrapper(id: Id, *args: P.args, **kwargs: P.kwargs) -> R:
        with namespace_context(id):
            return fn(*args, **kwargs)

    return wrapper


@no_example()
def server(
    fn: Callable[Concatenate[Inputs, Outputs, Session, P], R]
) -> Callable[Concatenate[str, P], R]:
    from .session import require_active_session, session_context

    def wrapper(id: Id, *args: P.args, **kwargs: P.kwargs) -> R:
        sess = require_active_session(None)
        child_sess = sess.make_scope(id)
        with session_context(child_sess):
            return fn(child_sess.input, child_sess.output, child_sess, *args, **kwargs)

    return wrapper



import asyncio
import random
from datetime import date

from shiny.express import render


@render.download(
    filename=lambda: f"新型-{date.today().isoformat()}-{random.randint(100, 999)}.csv"
)
async def downloadData():
    await asyncio.sleep(0.25)
    yield "one,two,three\n"
    yield "新,1,2\n"
    yield "型,4,5\n"



from shiny.express import input, render, ui

ui.page_opts(title="Selectize Inputs kitchensink")

fruits = ["Apple", "Banana", "Cherry", "Date", "Elderberry"]
fruits_dict = {
    "apple": "Apple",
    "banana": "Banana",
    "cherry": "Cherry",
    "date": "Date",
    "elderberry": "Elderberry",
}

fruits_grouped_dict = {
    "Citrus": {
        "Orange": "Sweet and tangy",
        "Lemon": "Zesty and refreshing",
        "Lime": "Bright and tart",
    },
    "Berries": {
        "Strawberry": "Juicy and sweet",
        "Blueberry": "Tiny and antioxidant-rich",
        "Raspberry": "Delicate and slightly tart",
    },
}


# Currently not wrapping in cards, as opening the selectize within a short card hides the selectize dropdown
ui.input_selectize("basic_selectize", "Default selectize", fruits)


@render.code
def basic_result_txt():
    return str(input.basic_selectize())


ui.input_selectize("selectize_with_label", "Selectize with label", fruits_dict)


@render.code
def selectize_with_label_txt():
    return str(input.selectize_with_label())


ui.input_selectize("multi_selectize", "Multiple Selectize", fruits, multiple=True)


@render.code
def multi_result_txt():
    return ", ".join(input.multi_selectize())


ui.input_selectize(
    "selectize_with_selected",
    "Selectize with selected",
    fruits,
    selected="Cherry",
)


@render.code
def selected_result_txt():
    return str(input.selectize_with_selected())


ui.input_selectize(
    "selectize_width_close_button",
    "Selectize with Custom Width and remove btn",
    fruits_grouped_dict,
    width="400px",
    remove_button=True,
)


@render.code
def selectize_width_close_button_txt():
    return str(input.selectize_width_close_button())



from __future__ import annotations

from shiny import App, Inputs, Outputs, Session, reactive, render, req, ui


def make_panel(letter: str) -> ui.AccordionPanel:
    return ui.accordion_panel(
        f"Section {letter}",
        f"Some narrative for section {letter}",
    )


items = [make_panel(letter) for letter in "ABCD"]

accordion = ui.accordion(*items, id="acc")
app_ui = ui.page_fluid(
    ui.tags.div(
        ui.input_action_button("toggle_b", "Open/Close B"),
        ui.input_action_button("open_all", "Open All"),
        ui.input_action_button("close_all", "Close All"),
        ui.input_action_button("alternate", "Alternate"),
        ui.input_action_button("toggle_efg", "Add/Remove EFG"),
        ui.input_action_button("toggle_updates", "Add/Remove Updates"),
        class_="d-flex",
    ),
    ui.output_text_verbatim("acc_txt", placeholder=True),
    accordion,
)


def server(input: Inputs, output: Outputs, session: Session) -> None:
    @reactive.calc
    def acc() -> list[str]:
        acc_val: list[str] | None = input.acc()
        if acc_val is None:
            acc_val = []
        return acc_val

    @reactive.effect
    def _():
        req(input.toggle_b())

        with reactive.isolate():
            if "Section B" in acc():
                ui.update_accordion_panel("acc", "Section B", show=False)
            else:
                ui.update_accordion_panel("acc", "Section B", show=True)

    @reactive.effect
    def _():
        req(input.open_all())
        ui.update_accordion("acc", show=True)

    @reactive.effect
    def _():
        req(input.close_all())
        ui.update_accordion("acc", show=False)

    has_efg = False
    has_alternate = True
    has_updates = False

    @reactive.effect
    def _():
        req(input.alternate())

        sections = [
            "updated_section_a" if has_updates else "Section A",
            "Section B",
            "Section C",
            "Section D",
        ]
        if has_efg:
            sections.extend(["Section E", "Section F", "Section G"])

        nonlocal has_alternate
        val = int(has_alternate)
        sections = [section for i, section in enumerate(sections) if i % 2 == val]
        ui.update_accordion("acc", show=sections)
        has_alternate = not has_alternate

    @reactive.effect
    def _():
        req(input.toggle_efg())

        nonlocal has_efg
        if has_efg:
            ui.remove_accordion_panel("acc", ["Section E", "Section F", "Section G"])
        else:
            ui.insert_accordion_panel("acc", make_panel("E"), "Section D")
            ui.insert_accordion_panel("acc", make_panel("F"), "Section E")
            ui.insert_accordion_panel("acc", make_panel("G"), "Section F")

        has_efg = not has_efg

    @reactive.effect
    def _():
        req(input.toggle_updates())

        nonlocal has_updates
        if has_updates:
            ui.update_accordion_panel(
                "acc",
                "updated_section_a",
                "Some narrative for section A",
                title="Section A",
                value="Section A",
                icon="",
            )
        else:
            with reactive.isolate():
                # print(acc())
                if "Section A" not in acc():
                    ui.notification_show("Opening Section A", duration=2)
                    ui.update_accordion_panel("acc", "Section A", show=True)
            ui.update_accordion_panel(
                "acc",
                "Section A",
                "Updated body",
                value="updated_section_a",
                title=ui.tags.h3("Updated title"),
                icon=ui.tags.div(
                    "Look! An icon! -->",
                    ui.HTML(
                        """\
                        <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-chat-right-quote" viewBox="0 0 16 16">
                          <path d="M2 1a1 1 0 0 0-1 1v8a1 1 0 0 0 1 1h9.586a2 2 0 0 1 1.414.586l2 2V2a1 1 0 0 0-1-1H2zm12-1a2 2 0 0 1 2 2v12.793a.5.5 0 0 1-.854.353l-2.853-2.853a1 1 0 0 0-.707-.293H2a2 2 0 0 1-2-2V2a2 2 0 0 1 2-2h12z"/>
                          <path d="M7.066 4.76A1.665 1.665 0 0 0 4 5.668a1.667 1.667 0 0 0 2.561 1.406c-.131.389-.375.804-.777 1.22a.417.417 0 1 0 .6.58c1.486-1.54 1.293-3.214.682-4.112zm4 0A1.665 1.665 0 0 0 8 5.668a1.667 1.667 0 0 0 2.561 1.406c-.131.389-.375.804-.777 1.22a.417.417 0 1 0 .6.58c1.486-1.54 1.293-3.214.682-4.112z"/>
                        </svg>
                        """
                    ),
                ),
            )

        has_updates = not has_updates

    @render.text
    def acc_txt():
        return f"input.acc(): {input.acc()}"


app = App(app_ui, server)
